Package: PTMods
Type: Package
Version: 1.1.0
Title: Managing Post-Translational Modifications in R
Authors@R: c(person("Laurent", "Gatto", role = "aut",
                    email = "laurent.gatto@uclouvain.be",
                    comment = c(ORCID = "0000-0002-1520-2268")),
             person("Sebastian", "Gibb", role = "aut",
                    email = "mail@sebastiangibb.de",
                    comment = c(ORCID = "0000-0001-7406-4443")),
             person("Guillaume", "Deflandre", role = "cre",
                    email = "guillaume.deflandre@uclouvain.be",
                    comment = c(ORCID = "0009-0008-1257-2416")))
Description: An interface to the community supported database for
             amino acid/protein modifications using mass
             spectrometry.
Depends: R (>= 4.5.0), methods
Suggests: xml2, testthat, knitr, BiocStyle, Biostrings
Enhances: PSMatch
License: GPL-3
VignetteBuilder: knitr
biocViews: Proteomics, MassSpectrometry
BugReports: https://github.com/RforMassSpectrometry/PTMods/issues
URL: https://github.com/RforMassSpectrometry/PTMods
RoxygenNote: 7.3.3
Encoding: UTF-8
git_url: https://git.bioconductor.org/packages/PTMods
git_branch: devel
git_last_commit: 08ea1e1
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:27:25 UTC; biocbuild
Author: Laurent Gatto [aut] (ORCID: <https://orcid.org/0000-0002-1520-2268>),
  Sebastian Gibb [aut] (ORCID: <https://orcid.org/0000-0001-7406-4443>),
  Guillaume Deflandre [cre] (ORCID:
    <https://orcid.org/0009-0008-1257-2416>)
Maintainer: Guillaume Deflandre <guillaume.deflandre@uclouvain.be>

# Generated by roxygen2: do not edit by hand

export(addFixedModifications)
export(addModifications)
export(addVariableModifications)
export(combineModifications)
export(convertAnnotation)
export(getCanonicalSequence)
export(getModificationsCounts)
importFrom(methods,is)
importFrom(stats,setNames)
importFrom(utils,combn)
importFrom(utils,data)
importFrom(utils,packageVersion)

#' Internal character id for unimod entries
#'
#' @param name character, unimod title
#' @param site character, unimod site
#' @param position character, unimod position
#' @param neutralLoss character, "0"/"1"
#' @return character
#' @noRd
.characterId <- function(name, site, position, neutralLoss) {
    if (!is.character(name))
        stop("'name' has to be a character.")
    if (!is.character(site))
        stop("'site' has to be a character.")
    if (!is.character(position))
        stop("'position' has to be a character.")
    if (!is.character(neutralLoss))
        stop("'neutralLoss' has to be a character.")
    id <- paste0(
        name, ":",
        ifelse(grepl("Protein", position), "P-", ""),
        site,
        ifelse(neutralLoss == "1", ":NL", "")
    )
    if (anyDuplicated(id)) {
        split(id, id) <- lapply(split(id, id), function(x) {
            if (length(x) > 1)
                paste0(x, ":", seq_along(x))
            else
                x
        })
    } 
    id
}

#' Internal function to query delta mass.
#'
#' @param xml xml_nodeset, <mod>
#' @return double
#' @noRd
.delta <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    node <- xml2::xml_find_first(xml, ".//umod:delta")
    xml2::xml_attrs(node)[c("avge_mass", "mono_mass", "composition")]
}

#' Internal function to turn unimod xml into a data.frame.
#'
#' @param xml xml_document (returned by .unimodDb)
#' @return data.frame
#' @noRd
.modifications <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    nodes <- xml2::xml_find_all(xml, "//umod:mod")
    u  <- lapply(nodes, function(n) {
        td <- c(.title(n), .delta(n))
        sp <- .specificity(n)
        m <- cbind(matrix(td, ncol=length(td), nrow=nrow(sp), byrow=TRUE,
                          dimnames=list(c(), names(td))), sp, NeutralLoss="0")
        nl <- .neutralLoss(n)
        if (!is.null(nl) && nrow(nl)) {
            nlrepl <- match(
                paste(nl[, "site"], nl[, "spec_group"]),
                paste(m[, "site"], m[, "spec_group"])
            )
            nl <- cbind(nl, NeutralLoss="1")
            mnr <- nrow(m)
            m <- rbind(m, m[nlrepl,,drop=FALSE])
            m[(mnr + 1L):nrow(m), colnames(nl)] <- nl
        }
        m
    })
    u <- do.call(rbind, u)
    m <- data.frame(
        Id=.characterId(
            u[, "title"], u[, "site"], u[, "position"], u[, "NeutralLoss"]
        ),
        UnimodId=as.numeric(u[, "record_id"]),
        Name=u[, "title"],
        Description=u[, "full_name"],
        Composition=u[, "composition"],
        AvgMass=as.numeric(u[, "avge_mass"]),
        MonoMass=as.numeric(u[, "mono_mass"]),
        Site=u[, "site"],
        Position=factor(u[, "position"]),
        Classification=factor(u[, "classification"]),
        SpecGroup=as.numeric(u[, "spec_group"]),
        NeutralLoss=as.logical(as.numeric(u[, "NeutralLoss"])),
        LastModified=u[, "date_time_modified"],
        Approved=as.logical(as.numeric(u[, "approved"])),
        Hidden=as.logical(as.numeric(u[, "hidden"])),
        stringsAsFactors=FALSE
    )
    rownames(m) <- m$Id
    m
}

#' Internal function to query neutral loss.
#'
#' @param xml xml_nodeset, <umod:specificity>
#' @return matrix
#' @noRd
.neutralLoss <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    nodes <- xml2::xml_find_all(xml, ".//umod:NeutralLoss")
    nl <- do.call(rbind, lapply(nodes, function(nd) {
        c(
            xml2::xml_attrs(xml2::xml_parent(nd))[c("site", "spec_group")],
            xml2::xml_attrs(nd)[c("avge_mass", "mono_mass", "composition")]
        )
    }))
    nl[nl[, "composition"] != "0",, drop=FALSE]
}

#' Internal function to query specificity.
#'
#' @param xml xml_nodeset, <mod>
#' @return matrix
#' @noRd
.specificity <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    nodes <- xml2::xml_find_all(xml, ".//umod:specificity")
    sp <- do.call(rbind, xml2::xml_attrs(nodes))
    sp[order(as.numeric(sp[, "spec_group"])),, drop=FALSE]
}

#' Internal function to query title/id/lastmod.
#'
#' @param xml xml_nodeset, <mod>
#' @return double
#' @noRd
.title <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    xml2::xml_attrs(xml)[c("record_id", "title", "full_name",
                           "date_time_modified", "approved")]
}

# may move to longtests/
# see: https://stat.ethz.ch/pipermail/bioc-devel/2017-November/012327.html

skip_if_not_installed("xml2")

xml <- PTMods:::.unimodDb()

test_that("aminoacids", {
    data(aminoacids, package="PTMods")
    attr(aminoacids, "Created") <- NULL ## ignore this attr
    expect_equal(PTMods:::.aminoacids(xml), aminoacids)
})

test_that("elements", {
    data(elements, package="PTMods")
    attr(elements, "Created") <- NULL ## ignore this attr
    expect_equal(PTMods:::.elements(xml), elements)
})

test_that("modifications", {
    data(modifications, package="PTMods")
    attr(modifications, "Created") <- NULL ## ignore this attr
    expect_equal(PTMods:::.modifications(xml), modifications)
})

skip_if_not_installed("xml2")
x <- xml2::read_xml(paste0(
    '<umod:unimod xmlns:umod="http://www.unimod.org/xmlns/schema/unimod_2">',
    '<umod:modifications>',
     '<umod:mod title="Acetyl" full_name="Acetylation" username_of_poster="unimod" group_of_poster="admin" date_time_posted="2002-08-19 19:17:11" date_time_modified="2008-02-15 05:20:02" approved="1" record_id="1">',
    '<umod:specificity hidden="1" site="S" position="Anywhere" classification="Post-translational" spec_group="4"/>',
     '<umod:specificity hidden="1" site="C" position="Anywhere" classification="Post-translational" spec_group="3"/>',
    '<umod:specificity hidden="0" site="N-term" position="Any N-term" classification="Multiple" spec_group="2">',
    '<umod:misc_notes>GIST acetyl light</umod:misc_notes>',
    '</umod:specificity>',
    '<umod:delta mono_mass="42.010565" avge_mass="42.0367" composition="H(2) C(2) O"> <umod:element symbol="H" number="2"/> <umod:element symbol="C" number="2"/> <umod:element symbol="O" number="1"/>',
    '</umod:delta>',
    '<umod:xref>',
    '<umod:text>14730666</umod:text>',
    '<umod:source>PubMed PMID</umod:source>',
    '<umod:url/>',
    '</umod:xref>',
    '<umod:xref>',
    '<umod:text>15350136</umod:text>',
    '<umod:source>PubMed PMID</umod:source>',
    '<umod:url/>',
    '</umod:xref>',
    '</umod:mod>',
    '</umod:modifications>',
    '</umod:unimod>'))
node <- xml2::xml_find_first(x, "//umod:mod[@record_id=\"1\"]")

sp <- cbind(
    hidden=c("0", "1", "1"),
    site=c("N-term", "C", "S"),
    position=c("Any N-term", rep("Anywhere", 2)),
    classification=c("Multiple", rep("Post-translational", 2)),
    spec_group=as.character(2:4)
)

test_that(".characterId", {
    expect_error(PTMods:::.characterId(1:10, "foo", "bar", "0"), "name")
    expect_error(PTMods:::.characterId("foo", 1:10, "bar", "0"), "site")
    expect_error(PTMods:::.characterId("foo", "bar", 1:10, "0"), "position")
    expect_error(PTMods:::.characterId("foo", "bar", "0", 1:10), "neutralLoss")
    expect_equal(
        PTMods:::.characterId(
            rep("Acetyl", 4),
            c("K", "S", "N-term", "N-term"),
            c("Anywhere", "Anywhere", "Any N-term", "Protein N-term"),
            c("0", "1", "0", "0")
        ),
        c("Acetyl:K", "Acetyl:S:NL", "Acetyl:N-term", "Acetyl:P-N-term")
    )
    expect_equal(
        PTMods:::.characterId(
            rep("Acetyl", 4),
            c("K", "S", "N-term", "N-term"),
            c("Anywhere", "Anywhere", "Any N-term", "Any N-term"),
            c("0", "1", "0", "0")
        ),
        c("Acetyl:K", "Acetyl:S:NL", "Acetyl:N-term:1", "Acetyl:N-term:2")
    )
})

test_that(".delta", {
    delta <- c(avge_mass="42.0367", mono_mass="42.010565",
               composition="H(2) C(2) O")
    expect_equal(PTMods:::.delta(node), delta)
})

test_that(".modifications", {
    m <- data.frame(Id=c("Acetyl:N-term", "Acetyl:C", "Acetyl:S"),
                    UnimodId=1, Name="Acetyl", Description="Acetylation",
                    Composition="H(2) C(2) O",
                    AvgMass=42.0367, MonoMass=42.010565,
                    Site=c("N-term", "C", "S"),
                    Position=factor(c("Any N-term", "Anywhere", "Anywhere")),
                    Classification=factor(c("Multiple", "Post-translational",
                                            "Post-translational")),
                    SpecGroup=2:4,
                    NeutralLoss=FALSE,
                    LastModified="2008-02-15 05:20:02",
                    Approved=TRUE, Hidden=c(FALSE, TRUE, TRUE),
                    stringsAsFactors=FALSE)
    rownames(m) <- m$Id
    expect_equal(PTMods:::.modifications(x), m)
})

test_that(".neutralLoss", {
    node <- xml2::read_xml(paste0(
'<umod:unimod xmlns:umod="http://www.unimod.org/xmlns/schema/unimod_2">',
'<umod:modifications>',
'<umod:mod title="Phospho" full_name="Phosphorylation" username_of_poster="unimod"',
'   group_of_poster="admin"',
'   date_time_posted="2002-08-19 19:17:11"',
'   date_time_modified="2011-11-25 10:55:54"',
'   approved="1"',
'   record_id="21">',
'<umod:specificity hidden="0" site="T" position="Anywhere" classification="Post-translational" spec_group="1">',
'<umod:NeutralLoss mono_mass="97.976896" avge_mass="97.9952" flag="false" composition="H(3) O(4) P">',
'<umod:element symbol="H" number="3"/>',
'<umod:element symbol="O" number="4"/>',
'<umod:element symbol="P" number="1"/>',
'</umod:NeutralLoss>',
'<umod:NeutralLoss mono_mass="0.000000" avge_mass="0.0000" flag="false" composition="0"/>',
'</umod:specificity>',
'<umod:delta mono_mass="79.966331" avge_mass="79.9799" composition="H O(3) P">',
'<umod:element symbol="H" number="1"/>',
'<umod:element symbol="O" number="3"/>',
'<umod:element symbol="P" number="1"/>',
'</umod:delta>',
'</umod:mod>',
'</umod:modifications>',
'</umod:unimod>'))
    nl <- matrix(c("T", "1", "97.9952", "97.976896", "H(3) O(4) P"), nrow=1,
                 dimnames=list(c(), c("site", "spec_group",
                                      "avge_mass", "mono_mass", "composition")))
    expect_equal(PTMods:::.neutralLoss(node), nl)
})

test_that(".specificity", {
    expect_equal(PTMods:::.specificity(node), sp)

    sp2 <- cbind(
        hidden="1",
        site="S",
        position="Anywhere",
        classification="Post-translational",
        spec_group="4"
    )
    node2 <- xml2::xml_find_first(xml2::read_xml(paste0(
'<umod:unimod xmlns:umod="http://www.unimod.org/xmlns/schema/unimod_2">',
'<umod:modifications>',
'<umod:mod title="Acetyl" full_name="Acetylation" username_of_poster="unimod" group_of_poster="admin" date_time_posted="2002-08-19 19:17:11" date_time_modified="2008-02-15 05:20:02" approved="1" record_id="1">',
'<umod:specificity hidden="1" site="S" position="Anywhere" classification="Post-translational" spec_group="4"/>',
'</umod:mod>',
'</umod:modifications>',
'</umod:unimod>')), "//umod:mod")
    expect_equal(PTMods:::.specificity(node2), sp2)
})

test_that(".title", {
    title <- c(record_id="1", title="Acetyl", full_name="Acetylation",
               date_time_modified="2008-02-15 05:20:02", approved="1")
    expect_equal(PTMods:::.title(node), title)
})

#' @title Combines consecutive modifications on amino acids in peptide
#' sequences.
#'
#' @param sequences `character`. A character vector of peptide sequences with
#' modifications in any format accepted by `convertAnnotation`.
#' Multiple modifications per amino acid are allowed.
#'
#' @return A character vector of the same length as `sequences` with
#' consecutive modifications on each amino acid summed into a single
#' modification. This summed modifications is written in delta mass format.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @examples
#' combineModifications("AT[+10][-5]K")
#' combineModifications("[+10]-[-5]-ATK")
#' combineModifications("AT[Phospho][UNIMOD:1]K")
#' combineModifications("EM[+15.9949][-10]EK[+42][-8]")
#' combineModifications(c("AT[+10][-5]K", "EM[+15.9949][-10]EK"))
#'
#' @export
combineModifications <- function(sequences) {
    sequences <- convertAnnotation(sequences)

    sapply(sequences, function(sequence) {
        sequence <- .combineTerminalMods(sequence)
        terms <- .extractTerminalMods(sequence)
        sequence <- terms$sequence

        masses <- .parseModifiedSequence(sequence)
        rx_residue <- "[A-Z](?:\\[(?:[GMURX]:)?[+-][0-9.]+\\])*"
        tokens <- regmatches(
            sequence,
            gregexpr(rx_residue, sequence, perl = TRUE)
        )[[1L]]
        aa <- substring(tokens, 1L, 1L)

        inner <- paste(
            ifelse(masses != 0,
                paste0(aa, "[", ifelse(masses > 0, "+", ""), masses, "]"),
                aa
            ),
            collapse = ""
        )

        .reattachTerminalMods(inner, terms$nterm, terms$cterm)
    }, USE.NAMES = FALSE)
}

#' @title Generates a numeric with positioned modifications
#'
#' @param sequence `character(1L)`. A peptide sequence that may have
#' modifications in delta mass format. Multiple modifications per
#' amino acid are allowed.
#'
#' @return `numeric()` of length equal to canonical sequence with specified
#' modifications.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @noRd
#'
#' @examples
#' PTMods:::.parseModifiedSequence("APEPT[+79.966]IDEK")
#' PTMods:::.parseModifiedSequence("[+304]-APEPTIDEK")
#' PTMods:::.parseModifiedSequence("EM[+15.9949]EVEES[+79.9663]PEK")
#' PTMods:::.parseModifiedSequence("APEPT[+79.966][+10]IDEK")
#' PTMods:::.parseModifiedSequence("EM[+15.9949][-10]EK")
.parseModifiedSequence <- function(sequence) {
    ## Match each amino acid followed by zero or more modification brackets
    rx_residue <- "[A-Z](?:\\[(?:[GMURX]:)?[+-][0-9.]+\\])*"
    ## Match delta masses: sign followed by digits, inside brackets
    rx_mass <- "[+-][0-9.]+"

    tokens <- regmatches(
        sequence,
        gregexpr(rx_residue, sequence, perl = TRUE)
    )[[1L]]

    masses <- vapply(tokens,
        function(token) {
            ## Remove leading amino acid character before extracting masses
            bracket_section <- substring(token, 2L)
            masses <- regmatches(
                bracket_section,
                gregexpr(rx_mass, bracket_section, perl = TRUE)
            )[[1L]]

            if (length(masses) == 0L) {
                return(0)
            }
            sum(as.double(masses))
        },
        numeric(1L),
        USE.NAMES = FALSE
    )

    ## Extract terminal modification values as attributes
    terms <- .extractTerminalMods(sequence)

    .extractTerminalMass <- function(term) {
        if (!nchar(term)) {
            return(NA_real_)
        }
        masses <- regmatches(term, gregexpr(rx_mass, term, perl = TRUE))[[1L]]
        sum(as.numeric(masses))
    }

    attr(masses, "Nterm") <- .extractTerminalMass(terms$nterm)
    attr(masses, "Cterm") <- .extractTerminalMass(terms$cterm)

    masses
}

#' @title Combine stacked terminal modifications into a single bracket.
#'
#' @param sequence `character(1L)`. A peptide sequence possibly containing
#' stacked N-terminal or C-terminal modifications.
#'
#' @return `character(1L)` with each terminal stack summed into a single
#' bracketed delta mass.
#'
#' @noRd
#'
#' @examples
#' .combineTerminalMods("[+304]-[-54]-ATK")
#' .combineTerminalMods("ATK-[+42]-[-8]")
#' .combineTerminalMods("[+304]-[-54]-ATK-[+42]-[-8]")
#' .combineTerminalMods("ATK")
.combineTerminalMods <- function(sequence) {
    rx_mass <- "[+-][0-9.]+"

    ## N-terminal: one or more [±x]- prefixes
    nterm_match <- regmatches(
        sequence,
        regexpr("^(\\[[^\\]]*\\]-)+", sequence, perl = TRUE)
    )

    if (length(nterm_match) && nchar(nterm_match) > 0L) {
        masses <- as.numeric(regmatches(
            nterm_match,
            gregexpr(rx_mass, nterm_match, perl = TRUE)
        )[[1L]])
        total <- sum(masses)
        sequence <- paste0(
            "[", ifelse(total > 0, "+", ""), total, "]-",
            substring(sequence, nchar(nterm_match) + 1L)
        )
    }

    ## C-terminal: one or more -[±x] suffixes
    cterm_match <- regmatches(
        sequence,
        regexpr("(-\\[[^\\]]*\\])+$", sequence, perl = TRUE)
    )

    if (length(cterm_match) && nchar(cterm_match) > 0L) {
        masses <- as.numeric(regmatches(
            cterm_match,
            gregexpr(rx_mass, cterm_match, perl = TRUE)
        )[[1L]])
        total <- sum(masses)
        sequence <- paste0(
            substring(sequence, 1L, nchar(sequence) - nchar(cterm_match)),
            "-[", ifelse(total > 0, "+", ""), total, "]"
        )
    }

    sequence
}

# Test combineModifications

test_that("combineModifications sums stacked delta mass modifications", {
    expect_equal(combineModifications("AT[+10][-5]K"), "AT[+5]K")
})

test_that("combineModifications combines stacked N-terminal modifications", {
    expect_equal(combineModifications("[+10]-[-5]-ATK"), "[+5]-ATK")
})

test_that("combineModifications converts name and unimod to delta mass", {
    result <- combineModifications("AT[Phospho][UNIMOD:1]K")
    expect_equal(result, "AT[+121.976896]K")
})

test_that("combineModifications handles multiple modified residues", {
    expect_equal(
        combineModifications("EM[+15.9949][-10]EK[+42][-8]"),
        "EM[+5.9949]EK[+34]"
    )
})

test_that("combineModifications handles vectorised input", {
    result <- combineModifications(c("AT[+10][-5]K", "EM[+15.9949][-10]EK"))
    expect_equal(result, c("AT[+5]K", "EM[+5.9949]EK"))
})

test_that("combineModifications handles unmodified sequences", {
    expect_equal(combineModifications("PEPTIDE"), "PEPTIDE")
})

test_that("combineModifications with single modification is unchanged", {
    expect_equal(combineModifications("AT[+10]K"), "AT[+10]K")
})

test_that("combineModifications combines stacked C-terminal mods", {
    expect_equal(combineModifications("ATK-[+42]-[-8]"), "ATK-[+34]")
})

test_that("combineModifications combines both terminal stacked mods", {
    expect_equal(
        combineModifications("[+304]-[-54]-ATK-[+42]-[-8]"),
        "[+250]-ATK-[+34]"
    )
})


# Test internal .parseModifiedSequence

test_that(".parseModifiedSequence parses single modification", {
    result <- PTMods:::.parseModifiedSequence("APEPT[+79.966]IDEK")
    expect_equal(
        as.numeric(result),
        c(0, 0, 0, 0, 79.966, 0, 0, 0, 0)
    )
})

test_that(".parseModifiedSequence parses stacked modifications", {
    result <- PTMods:::.parseModifiedSequence("APEPT[+79.966][+10]IDEK")
    expect_equal(
        as.numeric(result),
        c(0, 0, 0, 0, 89.966, 0, 0, 0, 0)
    )
})

test_that(".parseModifiedSequence parses mixed sign stacked modifications", {
    result <- PTMods:::.parseModifiedSequence("EM[+15.9949][-10]EK")
    expect_equal(as.numeric(result), c(0, 5.9949, 0, 0))
})

test_that(".parseModifiedSequence sets terminal attributes", {
    result <- PTMods:::.parseModifiedSequence("[+304]-APEPTIDEK")
    expect_equal(attr(result, "Nterm"), 304)
    expect_true(is.na(attr(result, "Cterm")))

    result2 <- PTMods:::.parseModifiedSequence("ATK-[+42]")
    expect_true(is.na(attr(result2, "Nterm")))
    expect_equal(attr(result2, "Cterm"), 42)
})

test_that(".parseModifiedSequence handles unmodified sequence", {
    result <- PTMods:::.parseModifiedSequence("PEPTIDE")
    expect_equal(as.numeric(result), rep(0, 7))
})


# Test internal .combineTerminalMods

test_that(".combineTerminalMods combines stacked N-terminal mods", {
    expect_equal(
        PTMods:::.combineTerminalMods("[+304]-[-54]-ATK"),
        "[+250]-ATK"
    )
})

test_that(".combineTerminalMods combines stacked C-terminal mods", {
    expect_equal(
        PTMods:::.combineTerminalMods("ATK-[+42]-[-8]"),
        "ATK-[+34]"
    )
})

test_that(".combineTerminalMods combines both terminal stacks", {
    expect_equal(
        PTMods:::.combineTerminalMods("[+304]-[-54]-ATK-[+42]-[-8]"),
        "[+250]-ATK-[+34]"
    )
})

test_that(".combineTerminalMods handles no stacked mods", {
    expect_equal(PTMods:::.combineTerminalMods("ATK"), "ATK")
    expect_equal(PTMods:::.combineTerminalMods("[+304]-ATK"), "[+304]-ATK")
    expect_equal(PTMods:::.combineTerminalMods("ATK-[+42]"), "ATK-[+42]")
})

#' Parse <aa> tag.
#' @param xml xml_document (returned by .unimodDb)
#' @return data.frame
#' @noRd
.aminoacids <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    aaNodes <- xml2::xml_find_all(xml, "//umod:amino_acids/umod:aa")
    aaAttrs <- xml2::xml_attrs(aaNodes)
    eeAttrs <- unlist(xml2::xml_attrs(
        xml2::xml_find_all(xml, "//umod:amino_acids/umod:aa/umod:element")
    ))

    aa <- do.call(rbind, aaAttrs)

    numbers <- as.integer(eeAttrs["number" == names(eeAttrs)])
    names(numbers) <- eeAttrs["symbol" == names(eeAttrs)]
    numbers <- split(numbers, rep.int(aa[, "title"], xml2::xml_length(aaNodes)))
    numbers[["-"]] <- integer()

    allElements <- unique(unlist(lapply(numbers, names)))

    em <- do.call(rbind, lapply(aa[, "title"], function(a) {
        e <- numbers[[a]]
        d <- setdiff(allElements, names(e))
        e[d] <- 0L
        e
    }))

    d <- data.frame(
        OneLetter=aa[, "title"],
        ThreeLetter=aa[, "three_letter"],
        FullName=aa[, "full_name"],
        AvgMass=as.double(aa[, "avge_mass"]),
        MonoMass=as.double(aa[, "mono_mass"]),
        row.names=aa[, "title"],
        stringsAsFactors=FALSE
    )
    cbind(d, em)
}

## TODO: save data.frame aminoacids as public available data set
## `data(aminoacids)`

test_that(".aminoacids", {
    skip_if_not_installed("xml2")
    x <- xml2::read_xml(paste0(
        '<umod:unimod xmlns:umod="http://www.unimod.org/xmlns/schema/unimod_2">',
        '<umod:amino_acids>',
        '<umod:aa title="A" three_letter="Ala" full_name="Alanine" mono_mass="71.037114" avge_mass="71.0779">',
        '<umod:element symbol="H" number="5"/>',
        '<umod:element symbol="C" number="3"/>',
        '<umod:element symbol="N" number="1"/>',
        '<umod:element symbol="O" number="1"/>',
        '</umod:aa>',
        '<umod:aa title="C" three_letter="Cys" full_name="Cysteine" mono_mass="103.009185" avge_mass="103.1429">',
        '<umod:element symbol="H" number="5"/>',
        '<umod:element symbol="C" number="3"/>',
        '<umod:element symbol="N" number="1"/>',
        '<umod:element symbol="O" number="1"/>',
        '<umod:element symbol="S" number="1"/>',
        '</umod:aa>',
        '</umod:amino_acids>',
        '</umod:unimod>'
    ))
    d <- data.frame(
        OneLetter=c("A", "C"),
        ThreeLetter=c("Ala", "Cys"),
        FullName=c("Alanine", "Cysteine"),
        AvgMass=c(71.0779, 103.1429),
        MonoMass=c(71.037114, 103.009185),
        H=c(5L, 5L),
        C=c(3L, 3L),
        N=c(1L, 1L),
        O=c(1L, 1L),
        S=c(0L, 1L),
        row.names=c("A", "C"),
        stringsAsFactors=FALSE
    )
    expect_equal(PTMods:::.aminoacids(x), d)
})

#' Internal function to create the unimod datasets.
#'
#' Used for its sideeffects
#'
#' @param path path to save the datasets
#'
#' @noRd
.createDataSets <- function(path="data") {
    xml <- .unimodDb()
    elements <- .elements(xml)
    aminoacids <- .aminoacids(xml)
    modifications <- .modifications(xml)
    attr(modifications, "Created") <-
        attr(elements, "Created") <-
        attr(aminoacids, "Created") <-
        date()
    save(elements, file=file.path(path, "elements.RData"),
         compress="xz")
    save(aminoacids, file=file.path(path, "aminoacids.RData"),
         compress="xz")
    save(modifications, file=file.path(path, "modifications.RData"),
         compress="xz")
}

#' Elements data set.
#'
#' `data.frame` of chemical elements from the unimod database.
#'
#' It was created as
#' follows:
#'
#' ```
#' PTMods:::.createDataSets()
#' ```
#'
#' @usage data("elements")
#' @format A `data.frame` with 4 columns (Name, FullName, AvgMass, MonoMass) for
#' the chemical elements.
#' @source Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
#' @examples
#' data(elements)
#' head(elements)
"elements"

#' Amminoacids data set.
#'
#' `data.frame` of aminoacids from the unimod database.
#'
#' It was created as
#' follows:
#'
#' ```
#' PTMods:::.createDataSets()
#' ```
#'
#' @usage data("aminoacids")
#' @format A `data.frame` with 11 columns (OneLetter, ThreeLetter, FullName,
#' AvgMass, MonoMass, H, C, N, O S, Se) for the aminoacids. The H/C/N/O/S/Se
#' columns contain the number of elements that build the aminoacid.
#' @source Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
#' @examples
#' data(aminoacids)
#' head(aminoacids)
"aminoacids"

#' Modifications data set.
#'
#' `data.frame` of modifications from the unimod database.
#'
#' It was created as
#' follows:
#'
#' ```
#' PTMods:::.createDataSets()
#' ```
#'
#' @usage data("modifications")
#' @format A `data.frame` with 15 columns (Id (created by
#' Name:(Position-)Site(:NeutralLoss) because unimod id is not unique),
#' UnimodId, Name, Description, AvgMass, MonoMass, Site, Position,
#' Classification, SpecGroup, NeutralLoss, LastModification, Approved,
#' Hidden) for the modifications.
#' @source Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
#' @examples
#' data(modifications)
#' head(modifications)
"modifications"

#' @title Add variable modifications
#'
#' @description Generates all possible combinations of variable modifications
#' for peptide sequences.
#'
#' The annotation style of modifications already present in `sequences` is
#' preserved as-is.  The annotation style of newly applied modifications
#' matches the style supplied in `variableModifications`: numeric values
#' produce deltaMass notation (e.g. `[+79.966331]`), while character values
#' are used verbatim (e.g. `[Phospho]` or `[UNIMOD:21]`).
#'
#' @param sequences Character vector. Peptide sequences that may have
#' modifications or not.
#'
#' @param variableModifications Named `numeric` or `character`. If a
#' `character` is given, values must be in UniMod name or UniMod ID format
#' (e.g. `"Phospho"`, `"UNIMOD:21"`). The annotation style of the values is
#' preserved in the output.
#' Specifies which variable modifications are used on which amino acids.
#' Supplying multiple modifications for the same amino acid is not supported;
#' call the function consecutively instead:
#' \preformatted{
#'   seq |>
#'     addVariableModifications(c(T = "+57.024")) |>
#'     addVariableModifications(c(T = "Phospho"))
#' }
#'
#' @param maxMods Numeric. Indicates how many modifications can be applied
#' at once.
#'
#' @return A named list of character vectors with all possible combinations
#' of modifications, one element per input sequence.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @importFrom utils combn
#'
#' @export
#'
#' @examples
#' addVariableModifications(
#'     "APGHKA",
#'     variableModifications = c(A = 4, K = 5, S = 8),
#'     maxMods = 2
#' )
#' addVariableModifications(
#'     "APGHKA",
#'     variableModifications = c(A = 4, K = 5, S = 8),
#'     maxMods = 3
#' )
#' addVariableModifications(
#'     c("ATK", "PAK"),
#'     variableModifications = c(T = "Phospho", A = "UNIMOD:1")
#' )
#' addVariableModifications(
#'     "A[+10]KT[-5]",
#'     variableModifications = c(A = -20, T = "Phospho")
#' )
#'
#' ## Apply two modifications to the same residue consecutively
#' "ATK" |>
#'     addVariableModifications(c(T = "+57.024")) |>
#'     addVariableModifications(c(T = "Phospho"))
addVariableModifications <- function(sequences,
                                     variableModifications = NULL,
                                     maxMods = Inf) {
    unlist(lapply(sequences, .addVariableModifications,
        variableModifications = variableModifications,
        maxMods = maxMods
    ), use.names = FALSE)
}

#' @title Generates list of possible combinations of modifications
#'
#' @param sequence Character. A peptide sequence that may have modifications or
#' not.
#'
#' @param variableModifications Named `numeric` or `character`. If a
#' `character` is given, it needs to be in unimodId or name format as
#' specified in `convertAnnotation`.
#' Specifies which variable modifications are used on which amino acids.
#' Supplying multiple modifications for the same amino acid is not supported;
#' call `addVariableModifications()` consecutively instead:
#' \preformatted{
#'   seq |>
#'     addVariableModifications(c(T = "+57.024")) |>
#'     addVariableModifications(c(T = "Phospho"))
#' }
#'
#' @param maxMods Numeric. Indicates how many modifications can be applied at
#' once.
#'
#' @return A `character()` vector with all possible combinations of
#' modifications.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @importFrom utils combn
#'
#' @noRd
#'
#' @examples
#' .addVariableModifications(
#'     "APGHKA",
#'     variableModifications = c(A = 4, K = 5, S = 8),
#'     maxMods = 2
#' )
#' .addVariableModifications(
#'     "APGHKA",
#'     variableModifications = c(A = 4, K = 5, S = 8),
#'     maxMods = 3
#' )
#' .addVariableModifications(
#'     "ATK",
#'     variableModifications = c(T = "Phospho", A = "UNIMOD:1")
#' )
#' .addVariableModifications(
#'     "A[+10]KT[-5]",
#'     variableModifications = c(A = -5, T = "Phospho")
#' )
#'
#' ## Apply two modifications to the same residue consecutively
#' "ATK" |>
#'     addVariableModifications(c(T = "+57.024")) |>
#'     addVariableModifications(c(T = "Phospho"))
.addVariableModifications <- function(sequence,
                                      variableModifications = NULL,
                                      maxMods = Inf) {
    if (is.na(sequence)) return(NA_character_)

    if (!is.null(variableModifications)) {
        mod_table <- table(names(variableModifications))

        if (!all(mod_table == 1)) {
            stop(
                "Applying multiple modifications to the same amino acid is not ",
                "supported.\n",
                "Please use addVariableModifications() consecutively instead:\n\n",
                "  'ATK' |>\n",
                "    addVariableModifications(c(T = '+57.024')) |>\n",
                "    addVariableModifications(c(T = 'Phospho'))"
            )
        }
    }

    if (!length(variableModifications) || maxMods <= 0) {
        return(sequence)
    }

    terms <- .extractTerminalMods(sequence)
    sequence <- terms$sequence

    sequence <- regmatches(
        sequence,
        gregexpr("[A-Z](\\[[^\\]]*\\])*", sequence, perl = TRUE)
    )[[1]]
    seqAA <- sub("\\[.*", "", sequence)

    modifiedAA <- names(variableModifications)

    varMods <- .resolveModsToStrings(variableModifications)
    names(varMods) <- modifiedAA

    modifiable_positions_var <- which(seqAA %in% modifiedAA)
    l <- length(modifiable_positions_var)
    max_mods <- min(maxMods, l)

    .mod <- function(cmb,
                     seq_split = sequence,
                     seq_aa = seqAA,
                     var_mods = varMods) {
        mStr <- setNames(rep(NA_character_, length(seq_split)), seq_split)
        vals <- unname(var_mods[seq_aa[cmb]])
        valid <- !is.na(vals)
        mStr[cmb[valid]] <- vals[valid]
        mStr
    }

    combos <- c(
        list(setNames(rep(NA_character_, length(sequence)), sequence)),
        unlist(
            lapply(seq_len(max_mods), function(n) {
                combn(modifiable_positions_var, n, FUN = .mod, simplify = FALSE)
            }),
            recursive = FALSE
        )
    )

    inner_sequences <- unique(vapply(combos, function(x) {
        paste(
            ifelse(!is.na(x),
                paste0(names(x), "[", x, "]"),
                names(x)
            ),
            collapse = ""
        )
    }, FUN.VALUE = character(1L), USE.NAMES = FALSE))

    vapply(inner_sequences, .reattachTerminalMods,
        nterm = terms$nterm, cterm = terms$cterm,
        FUN.VALUE = character(1L), USE.NAMES = FALSE
    )
}

# Test addVariableModifications

test_that("addVariableModifications generates all combos with maxMods", {
    result <- addVariableModifications("APGHKA",
        variableModifications = c(A = 4, K = 5, S = 8),
        maxMods = 2
    )
    ## Modifiable positions:
    ## 0 mods: APGHKA
    ## 1 mod:  A[+4]PGHKA, APGHK[+5]A, APGHKA[+4]
    ## 2 mods: A[+4]PGHK[+5]A, A[+4]PGHKA[+4], APGHK[+5]A[+4]
    expect_length(result, 7)
    expect_true("APGHKA" %in% result)
    expect_true("A[+4]PGHKA" %in% result)
    expect_true("APGHK[+5]A" %in% result)
    expect_true("APGHKA[+4]" %in% result)
    expect_true("A[+4]PGHK[+5]A" %in% result)
    expect_true("A[+4]PGHKA[+4]" %in% result)
    expect_true("APGHK[+5]A[+4]" %in% result)
})

test_that("addVariableModifications generates all combos with maxMods = 3", {
    result <- addVariableModifications("APGHKA",
        variableModifications = c(A = 4, K = 5, S = 8),
        maxMods = 3
    )
    ## Same as maxMods=2 plus the 3-mod combo:
    ## A[+4]PGHK[+5]A[+4]
    expect_length(result, 8)
    expect_true("A[+4]PGHK[+5]A[+4]" %in% result)
})

test_that("addVariableModifications handles character modifications preserving style", {
    result <- addVariableModifications("ATK",
        variableModifications = c(T = "Phospho", A = "UNIMOD:1")
    )
    ## Annotation style of the parameter is preserved in the output:
    ## 0 mods: ATK
    ## 1 mod:  A[UNIMOD:1]TK, AT[Phospho]K
    ## 2 mods: A[UNIMOD:1]T[Phospho]K
    expect_length(result, 4)
    expect_true("ATK" %in% result)
    expect_true("A[UNIMOD:1]TK" %in% result)
    expect_true("AT[Phospho]K" %in% result)
    expect_true("A[UNIMOD:1]T[Phospho]K" %in% result)
})

test_that("addVariableModifications preserves existing annotation style while adding character mods", {
    result <- addVariableModifications("A[+42]TK",
        variableModifications = c(T = "Phospho")
    )
    ## Existing [+42] stays in deltaMass; new mod uses name style as supplied
    expect_true("A[+42]TK" %in% result)
    expect_true("A[+42]T[Phospho]K" %in% result)
    expect_length(result, 2)
})

test_that("addVariableModifications handles vectorised input", {
    result <- addVariableModifications(c("ATK", "PAK"),
        variableModifications = c(T = 10, A = 5)
    )
    ## ATK: modifiable A(1), T(2) → 4 combos
    ## PAK: modifiable A(2) → 2 combos
    expect_length(result, 6)
})

test_that("addVariableModifications preserves existing modifications", {
    result <- addVariableModifications("A[+10]KT[-5]",
        variableModifications = c(A = -5, T = 8)
    )
    ## Modifiable: position 1 (A), position 3 (T)
    expect_true("A[+10]KT[-5]" %in% result)
    expect_length(result, 4)
})

test_that("addVariableModifications respects maxMods = 1", {
    result <- addVariableModifications("ATSK",
        variableModifications = c(A = 4, T = 5, S = 8),
        maxMods = 1
    )
    expect_length(result, 4)
    expect_true("ATSK" %in% result)
    expect_true("A[+4]TSK" %in% result)
    expect_true("AT[+5]SK" %in% result)
    expect_true("ATS[+8]K" %in% result)
})

test_that("addVariableModifications with no matching residues returns original", {
    result <- addVariableModifications("GGG",
        variableModifications = c(A = 4, S = 8)
    )
    expect_equal(result, "GGG")
})

test_that("addVariableModifications preserves terminal modifications", {
    result <- addVariableModifications("[+304]-ATK",
        variableModifications = c(A = 5)
    )
    expect_true("[+304]-ATK" %in% result)
    expect_true("[+304]-A[+5]TK" %in% result)
    expect_length(result, 2)
})

test_that("addVariableModifications returns original if maxMods = 0", {
    result <- addVariableModifications("ATSK",
        variableModifications = c(A = 4, T = 5),
        maxMods = 0
    )
    expect_equal(result, "ATSK")
})

test_that("addVariableModifications applies two mods to same residue consecutively", {
    result <- "ATK" |>
        addVariableModifications(c(T = "+57.024")) |>
        addVariableModifications(c(T = "Phospho"))
    ## All four combinations: unmodified, first mod only, second mod only,
    ## and both mods stacked on T
    expect_length(result, 4)
    expect_true("ATK" %in% result)
    expect_true("AT[+57.024]K" %in% result)
    expect_true("AT[Phospho]K" %in% result)
    expect_true("AT[+57.024][Phospho]K" %in% result)
})

test_that("addVariableModifications returns original with NULL mods", {
    result <- addVariableModifications("ATSK",
        variableModifications = NULL
    )
    expect_equal(result, "ATSK")
})

test_that("addVariableModifications preserves NAs in input", {
    result <- addVariableModifications(c(NA, "ATK"))
    expect_true(is.na(result[[1]]))
    expect_equal(result[[2]], "ATK")
})

#' character to composition
#'
#' @param x character
#' @return named numeric
#' @noRd
.character2composition <- function(x) {
    if (!is.character(x)) {
        stop("'x' must be a 'character'")
    }
    s <- strsplit(x, "[[:space:]]")[[1L]]
    nms <- gsub("[^[:alpha:]]+", "", s)
    v <- as.integer(gsub("[^[:digit:]]+", "", s))
    v[is.na(v)] <- 1L
    names(v) <- nms
    v
}

#' Composition to character
#'
#' @param x numeric, named
#' @return character
#' @noRd
.composition2character <- function(x) {
    if (!is.numeric(x)) {
        stop("'x' must be a 'numeric'")
    }
    if (is.null(names(x))) {
        stop("'x' must be a named 'numeric'")
    }
    paste0(names(x), "(", x, ")", collapse=" ")
}

#' Splits a string into a character vector
#'
#' @param x character, (AAString), (AAStringSet)
#' @return list of single characters
#' @importFrom methods is
#' @noRd
.string2character <- function(x) {
    stopifnot(
        is.character(x) ||
        class(x) %in% c("AAString", "AAStringSet", "AAStringSetList")
    )
    if (is(x, "AAStringSetList")) {
        x <- unlist(x)
    }
    ## if we always apply as.character the names of original characters are
    ## removed
    if (!is.character(x)) {
        x <- as.character(x)
    }
    strsplit(x, split=character(), fixed=TRUE)
}

test_that("getModificationsCounts unifies mixed annotation styles to UniMod name", {
    ## [+79.966] and [Phospho] both resolve to "Phospho"
    seqs <- c("T[Phospho]K", "S[+79.966]PEK")
    result <- getModificationsCounts(seqs)
    expect_equal(result, c(Phospho = 2L))
})

test_that("getModificationsCounts returns empty vector when no modifications", {
    result <- getModificationsCounts("PEPTIDE")
    expect_equal(result, setNames(integer(0L), character(0L)))
})

test_that("getModificationsCounts counts distinct modifications across sequences", {
    seqs <- c(
        "AC[Carbamidomethyl]TK",
        "M[Oxidation]EVNES[Phospho]PEK",
        "S[+79.966]PEK"
    )
    result <- getModificationsCounts(seqs)
    expect_equal(result[["Carbamidomethyl"]], 1L)
    expect_equal(result[["Oxidation"]], 1L)
    ## [Phospho] and [+79.966] should both count as Phospho
    expect_equal(result[["Phospho"]], 2L)
})

test_that(".composition2character", {
    expect_error(PTMods:::.composition2character("foo"), "must be a 'numeric'")
    expect_error(PTMods:::.composition2character(1:3), "must be a named")
    expect_equal(PTMods:::.composition2character(c(H=1, P=1, O=4)),
                 "H(1) P(1) O(4)")
})

test_that(".character2composition", {
    expect_error(PTMods:::.character2composition(1:3), "must be a 'character'")
    expect_equal(PTMods:::.character2composition("H(1) P(1) O(4)"),
                 c(H=1, P=1, O=4))
    expect_equal(PTMods:::.character2composition("H P O(4)"), c(H=1, P=1, O=4))
})

test_that(".string2character", {
    expect_error(PTMods:::.string2character(1:3))
    expect_equal(PTMods:::.string2character(c(a="ABC", b="DEF")),
                 list(a=c("A", "B", "C"), b=c("D", "E", "F")))
    skip_if_not_installed("Biostrings")
    expect_equal(PTMods:::.string2character(
        Biostrings::AAStringSet(c(a="ABC", b="DEF"))
    ), list(a=c("A", "B", "C"), b=c("D", "E", "F")))
})

utils::globalVariables("modifications")

#' @name convertAnnotation
#'
#' @aliases convertAnnotation .convertAnnotation
#'
#' @title Convert sequences from one annotation style to another
#'
#' @description
#' Converts modifications between different annotation formats for multiple
#' sequences at once. See the details and examples sections for more
#' information.  The annotation styles are inferred from the `modifications`
#' dataframe (see `?modifications`).
#'
#' @param x Character vector with peptide sequences in ProForma format
#'
#' @param convertToStyle Character string specifying target format. Options:
#'   "deltaMass", "unimodId", "name"
#'
#' @param massTolerance Numeric mass tolerance in Daltons for matching
#'   modifications (default: 0.01). Used when converting from deltaMass.
#'
#' @param verbose `logical(1)`. If `FALSE`, warnings about unrecognised
#'   modifications are silenced (default: `TRUE`).
#'
#' @return Character vector with the sequences in the target annotation format
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @importFrom utils data
#'
#' @details
#' The function handles three main conversion scenarios:
#'
#' \itemize{
#'   \item Name to deltaMass: \code{"M[Oxidation]PEPTIDE"} -> \code{"M[+15.994915]PEPTIDE"}
#'   \item Name to unimodId: \code{"M[Oxidation]PEPTIDE"} -> \code{"M[UNIMOD:35]PEPTIDE"}
#'   \item deltaMass to name: \code{"M[+15.995]PEPTIDE"} -> \code{"M[Oxidation]PEPTIDE"}
#'   \item deltaMass to unimodId: \code{"M[+15.995]PEPTIDE"} -> \code{"M[UNIMOD:35]PEPTIDE"}
#'   \item unimodId to name: \code{"M[UNIMOD:35]PEPTIDE"} -> \code{"M[Oxidation]PEPTIDE"}
#'   \item unimodId to deltaMass: \code{"M[UNIMOD:35]PEPTIDE"} -> \code{"M[+15.994915]PEPTIDE"}
#' }
#'
#' @examples
#' # Convert sequence from name to delta mass
#' convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "deltaMass")
#' # Result: "M[+15.994915]PEPTIDE"
#'
#' # Name to Unimod ID
#' convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "unimodId")
#' # Result: "M[UNIMOD:35]PEPTIDE"
#'
#' # Delta mass to name
#' convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "name")
#' # Result: "M[Oxidation]PEPTIDE"
#'
#' # Multiple modifications
#' convertAnnotation("M[Oxidation]EVNES[Phospho]PEK", convertToStyle = "deltaMass")
#' # Result: "M[+15.994915]EVNES[+79.966331]PEK"
#'
#' # Convert multiple sequences from name to delta mass
#' sequences <- c("M[Oxidation]PEPTIDE", "EVNES[Phospho]PEK", "PEPTIDE")
#' convertAnnotation(sequences, convertToStyle = "deltaMass")
#' # Result: c("M[+15.994915]PEPTIDE", "EVNES[+79.966331]PEK", "PEPTIDE")
#'
#' # Convert from delta mass to name
#' sequences <- c("M[+15.995]PEPTIDE", "S[+79.966]EQUENCE")
#' convertAnnotation(sequences, convertToStyle = "name")
#' # Result: c("M[Oxidation]PEPTIDE", "S[Phospho]EQUENCE")
#'
#' # Convert to Unimod IDs
#' sequences <- c("M[Oxidation]PEPTIDE", "C[Carbamidomethyl]PEPTIDE")
#' convertAnnotation(sequences, convertToStyle = "unimodId")
#' # Result: c("M[UNIMOD:35]PEPTIDE", "C[UNIMOD:4]PEPTIDE")
#'
#' @export
convertAnnotation <- function(x,
                              convertToStyle = c("deltaMass", "unimodId", "name"),
                              massTolerance = 0.01,
                              verbose = TRUE) {

    # Validate inputs
    if (!is.character(x)) {
        stop("x must be a character vector")
    }

    convertToStyle <- match.arg(convertToStyle)

    # Load modifications data once into the local environment
    data(modifications, envir = environment(), package = "PTMods")

    # Prepare unimodData once for all sequences
    unimodData <- modifications[!modifications$NeutralLoss,
                                c("UnimodId", "Name", "MonoMass")]
    unimodData <- unimodData[!duplicated(unimodData$Name), ]

    # Apply .convertAnnotation to each element
    vapply(x,
           function(seq) .convertAnnotation(seq,
                                           convertToStyle = convertToStyle,
                                           massTolerance = massTolerance,
                                           unimodData = unimodData,
                                           verbose = verbose),

           character(1),
           USE.NAMES = FALSE)
}


#' Converter for a character of length(1L)
#'
#' @noRd
.convertAnnotation <- function(x,
                              convertToStyle = c("deltaMass", "unimodId", "name"),
                              massTolerance = 0.01,
                              unimodData = NULL,
                              verbose = TRUE) {

    if (!is.character(x) || length(x) != 1L) {
        stop("x must be a single character string")
    }

    convertToStyle <- match.arg(convertToStyle)

    if (is.null(unimodData)) {
        # Load modifications data once into the local environment
        data(modifications, envir = environment(), package = "PTMods")

        # Prepare unimodData once for all sequences
        unimodData <- modifications[!modifications$NeutralLoss,
            c("UnimodId", "Name", "MonoMass")]

        unimodData <- unimodData[!duplicated(unimodData$Name), ]
    }

    # Find all modifications in the sequence
    # Pattern matches: [content] where content is not empty
    pattern <- "\\[([^]]+)\\]"

    # Extract all modifications
    matches <- gregexpr(pattern, x)
    mod_strings <- regmatches(x, matches)[[1]]

    if (length(mod_strings) == 0L) {
        # No modifications found, return unchanged
        return(x)
    }

    # Extract content within brackets
    modContents <- gsub("\\[|\\]", "", mod_strings)

    # Convert each modification
    result <- x
    for (i in seq_along(modContents)) {
        modContent <- modContents[i]

        # Determine the input type
        inputType <- .detectModificationType(modContent)

        # Skip if already in target format
        if (inputType == convertToStyle) {
            next
        }

        # Perform conversion
        converted <- .convertModificationFormat(
            modContent = modContent,
            inputType = inputType,
            outputType = convertToStyle,
            unimodData = unimodData,
            massTolerance = massTolerance,
            verbose = verbose
        )

        # Replace in the result string if conversion successful
        if (!is.null(converted)) {
            old_string <- mod_strings[i]
            new_string <- paste0("[", converted, "]")
            result <- sub(old_string, new_string, result, fixed = TRUE)
        }
    }

    return(result)
}


#' Detect modification type from string
#'
#' Determines whether a modification string represents a name, mass shift,
#' or Unimod ID.
#'
#' @param mod_string Character string with modification (without brackets)
#'
#' @return Character string: "name", "deltaMass", or "unimodId"
#'
#' @examples
#' # Detect Unimod ID format
#' PTMods:::.detectModificationType("UNIMOD:35")
#' # Result: "unimodId"
#'
#' # Detect delta mass format
#' PTMods:::.detectModificationType("+15.995")
#' # Result: "deltaMass"
#'
#' # Detect name format
#' PTMods:::.detectModificationType("Oxidation")
#' # Result: "name"
#'
#' @noRd
.detectModificationType <- function(mod_string) {
    # Check for Unimod ID (i.e. UNIMOD:35 or U:35)
    if (grepl("^(UNIMOD|U):[0-9]+$", mod_string, ignore.case = TRUE)) {
        return("unimodId")
    }

    # Check for mass shift (+15.995 or -17.026)
    if (grepl("^[+-]?[0-9]+(\\.[0-9]+)?$", mod_string)) {
        return("deltaMass")
    }

    # Otherwise assume it's a name
    return("name")
}


#' Convert modification between formats
#'
#' Converts a single modification between different annotation formats using
#' the Unimod database.
#'
#' @param modContent Character string with modification content (without brackets)
#'
#' @param inputType Character string: "name", "deltaMass", or "unimodId"
#'
#' @param outputType Character string: "name", "deltaMass", or "unimodId"
#'
#' @param unimodData Data.frame with Unimod data
#'
#' @param massTolerance Numeric mass tolerance in Daltons
#'
#' @return Character string with converted modification, or NULL if conversion failed
#'
#' @examples
#' # Load Unimod data
#' data(modifications)
#' unimodData <- modifications[!modifications$NeutralLoss,
#'                                      c("UnimodId", "Name", "MonoMass")]
#' unimodData <- unimodData[!duplicated(unimodData$Name), ]
#'
#' # Convert from name to delta mass
#' PTMods:::.convertModificationFormat("Oxidation", "name", "deltaMass",
#'                           unimodData, 0.01)
#' # Result: "+15.994915"
#'
#' # Convert from delta mass to name
#' PTMods:::.convertModificationFormat("+15.995", "deltaMass", "name",
#'                           unimodData, 0.01)
#' # Result: "Oxidation"
#'
#' @noRd
.convertModificationFormat <- function(modContent, inputType, outputType,
                                       unimodData, massTolerance,
                                       verbose = TRUE) {

    # Strategy: always convert to intermediate representation first
    # inputType → intermediate (UnimodId + Mass + Name) → outputType

    intermediate <- NULL

    # Step 1: Convert input to intermediate representation
    if (inputType == "name") {
        # Look up by name
        intermediate <- .lookupByName(modContent, unimodData)

    } else if (inputType == "deltaMass") {
        # Extract mass value
        mass_value <- as.numeric(modContent)
        # Look up by mass
        intermediate <- .lookupByMass(mass_value, unimodData, massTolerance)

    } else if (inputType == "unimodId") {
        # Extract Unimod ID
        unimodId <- .extractUnimodId(modContent)
        # Look up by ID
        intermediate <- .lookupByUnimodId(unimodId, unimodData)
    }

    # Check if lookup was successful
    if (is.null(intermediate)) {
        if (verbose) {
        warning(paste0(
            "Could not find Unimod entry for modification ",
            modContent, ", see `?modifications`"))
        }
        return(NULL)
    }

    # Step 2: Convert intermediate to output format
    if (outputType == "name") {
        return(intermediate$name)

    } else if (outputType == "deltaMass") {
        # Format mass with appropriate precision (by default 6 decimals)
        return(.formatMass(intermediate$mass))

    } else if (outputType == "unimodId") {
        return(paste0("UNIMOD:", intermediate$unimodId))
    }

    return(NULL)
}


#' Look Up Modification by Name
#'
#' Finds a modification in the Unimod database by its name.
#'
#' @param name Character string with modification name
#'
#' @param unimodData Data.frame with Unimod data
#'
#' @return List with unimodId, name, and mass, or NULL if not found
#'
#' @examples
#' # Load Unimod data
#' unimodData <- PTMods::modifications[!PTMods::modifications$NeutralLoss,
#'                                      c("UnimodId", "Name", "MonoMass")]
#' unimodData <- unimodData[!duplicated(unimodData$Name), ]
#'
#' # Look up by name
#' .lookupByName("Oxidation", unimodData)
#' # Result: list(unimodId = 35, name = "Oxidation", mass = 15.994915)
#'
#' # Look up non-existent modification
#' .lookupByName("NonExistent", unimodData)
#' # Result: NULL
#'
#' @noRd
.lookupByName <- function(name, unimodData) {
    # Match by name (exact match)
    match_idx <- which(unimodData$Name == name)

    if (length(match_idx) == 0L) {
        return(NULL)
    }

    # Return the first match
    list(
        unimodId = unimodData$UnimodId[match_idx[1]],
        name = unimodData$Name[match_idx[1]],
        mass = unimodData$MonoMass[match_idx[1]]
    )
}


#' Look Up Modification by Mass
#'
#' Finds a modification in the Unimod database by its monoisotopic mass.
#'
#' @param mass Numeric mass value
#'
#' @param unimodData Data.frame with Unimod data
#'
#' @param tolerance Numeric mass tolerance in Daltons
#'
#' @return List with unimodId, name, and mass, or NULL if not found
#'
#' @examples
#' # Load Unimod data
#' unimodData <- PTMods::modifications[!PTMods::modifications$NeutralLoss,
#'                                      c("UnimodId", "Name", "MonoMass")]
#' unimodData <- unimodData[!duplicated(unimodData$Name), ]
#'
#' # Look up by mass with tolerance
#' .lookupByMass(15.995, unimodData, 0.01)
#' # Result: list(unimodId = 35, name = "Oxidation", mass = 15.994915)
#'
#' # Look up with no match
#' .lookupByMass(999.999, unimodData, 0.01)
#' # Result: NULL
#'
#' @noRd
.lookupByMass <- function(mass, unimodData, tolerance) {
    # Match by mass within tolerance
    mass_diff <- abs(unimodData$MonoMass - mass)
    match_idx <- which(mass_diff <= tolerance)

    if (length(match_idx) == 0L) {
        return(NULL)
    }

    # If multiple matches, return the closest one
    if (length(match_idx) > 1L) {
        closest_idx <- match_idx[which.min(mass_diff[match_idx])]
        match_idx <- closest_idx
    }

    # Return the best match
    list(
        unimodId = unimodData$UnimodId[match_idx[1]],
        name = unimodData$Name[match_idx[1]],
        mass = unimodData$MonoMass[match_idx[1]]
    )
}


#' Look Up Modification by Unimod ID
#'
#' Finds a modification in the Unimod database by its Unimod ID.
#'
#' @param unimodId Numeric Unimod ID
#'
#' @param unimodData Data.frame with Unimod data
#'
#' @return List with unimodId, name, and mass, or NULL if not found
#'
#' @examples
#' # Load Unimod data
#' unimodData <- PTMods::modifications[!PTMods::modifications$NeutralLoss,
#'                                      c("UnimodId", "Name", "MonoMass")]
#' unimodData <- unimodData[!duplicated(unimodData$Name), ]
#'
#' # Look up by Unimod ID
#' .lookupByUnimodId(35, unimodData)
#' # Result: list(unimodId = 35, name = "Oxidation", mass = 15.994915)
#'
#' # Look up non-existent ID
#' .lookupByUnimodId(99999, unimodData)
#' # Result: NULL
#'
#' @noRd
.lookupByUnimodId <- function(unimodId, unimodData) {
    # Match by UnimodId
    match_idx <- which(unimodData$UnimodId == unimodId)

    if (length(match_idx) == 0L) {
        return(NULL)
    }

    # Return the match
    list(
        unimodId = unimodData$UnimodId[match_idx[1]],
        name = unimodData$Name[match_idx[1]],
        mass = unimodData$MonoMass[match_idx[1]]
    )
}


#' Extract Unimod ID from String
#'
#' Extracts the numeric Unimod ID from strings like "UNIMOD:35" or "U:35".
#'
#' @param unimod_string Character string with Unimod ID
#'
#' @return Numeric Unimod ID
#'
#' @examples
#' # Extract from full format
#' .extractUnimodId("UNIMOD:35")
#' # Result: 35
#'
#' # Extract from short format
#' .extractUnimodId("U:35")
#' # Result: 35
#'
#' # Case insensitive
#' .extractUnimodId("unimod:35")
#' # Result: 35
#'
#' @noRd
.extractUnimodId <- function(unimod_string) {
    # Extract the number after "UNIMOD:" or "U:"
    id_str <- sub("^(UNIMOD|U):", "", unimod_string, ignore.case = TRUE)
    as.integer(id_str)
}


#' Format Mass Value
#'
#' Formats a mass value as a string with appropriate sign and precision.
#'
#' @param mass Numeric mass value
#'
#' @param digits Integer number of decimal places (default: 6)
#'
#' @return Character string with formatted mass (e.g., "+15.994915")
#'
#' @examples
#' # Format positive mass
#' .formatMass(15.994915)
#' # Result: "+15.994915"
#'
#' # Format negative mass
#' .formatMass(-17.026549)
#' # Result: "-17.026549"
#'
#' # Custom precision
#' .formatMass(15.994915, digits = 3)
#' # Result: "+15.995"
#'
#' @noRd
.formatMass <- function(mass, digits = 6) {
    # Format with sign
    if (mass >= 0) {
        paste0("+", round(mass, digits))
    } else {
        as.character(round(mass, digits))
    }
}

## Test main convertAnnotation function

data(modifications)


test_that("convertAnnotation converts name to delta_mass", {
    # Single modification
    expect_equal(convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "deltaMass"),
                 "M[+15.994915]PEPTIDE")

    # Multiple modifications
    expect_equal(convertAnnotation("M[Oxidation]EVNES[Phospho]PEK",
                                   convertToStyle = "deltaMass"),
                 "M[+15.994915]EVNES[+79.966331]PEK")

    # No modifications
    expect_equal(convertAnnotation("PEPTIDE", convertToStyle = "deltaMass"),
                 "PEPTIDE")
})

test_that("convertAnnotation converts name to unimod_id", {
    # Single modification
    expect_equal(convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "unimodId"),
                 "M[UNIMOD:35]PEPTIDE")

    # Multiple modifications
    expect_equal(convertAnnotation("C[Carbamidomethyl]PEPTIDE",
                                   convertToStyle = "unimodId"),
                 "C[UNIMOD:4]PEPTIDE")
})

test_that("convertAnnotation converts delta_mass to name", {
    # Single modification
    expect_equal(convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "name"),
                 "M[Oxidation]PEPTIDE")

    # Multiple modifications
    expect_equal(convertAnnotation("S[+79.966]EQUENCE", convertToStyle = "name"),
                 "S[Phospho]EQUENCE")
})

test_that("convertAnnotation converts delta_mass to unimod_id", {
    expect_equal(convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "unimodId"),
                 "M[UNIMOD:35]PEPTIDE")

    expect_equal(convertAnnotation("S[+79.966]PEPTIDE", convertToStyle = "unimodId"),
                 "S[UNIMOD:21]PEPTIDE")
})

test_that("convertAnnotation converts unimod_id to name", {
    expect_equal(convertAnnotation("M[UNIMOD:35]PEPTIDE", convertToStyle = "name"),
                 "M[Oxidation]PEPTIDE")

    expect_equal(convertAnnotation("C[UNIMOD:4]PEPTIDE", convertToStyle = "name"),
                 "C[Carbamidomethyl]PEPTIDE")
})

test_that("convertAnnotation converts unimod_id to delta_mass", {
    expect_equal(convertAnnotation("M[UNIMOD:35]PEPTIDE",
                                   convertToStyle = "deltaMass"),
                 "M[+15.994915]PEPTIDE")

    expect_equal(convertAnnotation("S[UNIMOD:21]PEPTIDE",
                                   convertToStyle = "deltaMass"),
                 "S[+79.966331]PEPTIDE")
})

test_that("convertAnnotation handles vectorized input", {
    sequences <- c("M[Oxidation]PEPTIDE", "EVNES[Phospho]PEK", "PEPTIDE")
    result <- convertAnnotation(sequences, convertToStyle = "deltaMass")

    expect_length(result, 3)
    expect_equal(result[1], "M[+15.994915]PEPTIDE")
    expect_equal(result[2], "EVNES[+79.966331]PEK")
    expect_equal(result[3], "PEPTIDE")

    # No names should be preserved
    expect_null(names(result))
})

test_that("convertAnnotation respects mass_tolerance", {
    # With default tolerance (0.01), should match
    expect_equal(convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "name"),
                 "M[Oxidation]PEPTIDE")

    # With very strict tolerance, may not match
    expect_warning(
        result <- convertAnnotation("M[+16.1]PEPTIDE",
                                   convertToStyle = "name",
                                   massTolerance = 0.01),
        "Could not find Unimod entry"
    )
    expect_equal(result, "M[+16.1]PEPTIDE")
})

test_that("convertAnnotation handles sequences already in target format", {
    # Already in delta_mass format, converting to delta_mass
    expect_equal(convertAnnotation("M[+15.994915]PEPTIDE",
                                   convertToStyle = "deltaMass"),
                 "M[+15.994915]PEPTIDE")

    # Already in name format, converting to name
    expect_equal(convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "name"),
                 "M[Oxidation]PEPTIDE")

    # Already in unimod_id format, converting to unimod_id
    expect_equal(convertAnnotation("M[UNIMOD:35]PEPTIDE",
                                   convertToStyle = "unimodId"),
                 "M[UNIMOD:35]PEPTIDE")
})

test_that("convertAnnotation input validation", {
    # Non-character input
    expect_error(convertAnnotation(123, convertToStyle = "deltaMass"),
                 "x must be a character vector")

    expect_error(convertAnnotation(list("M[Oxidation]PEPTIDE"),
                                   convertToStyle = "deltaMass"),
                 "x must be a character vector")

    # Invalid convert_to argument
    expect_error(convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "invalid"),
                 "'arg' should be one of")
})

test_that("convertAnnotation warns on unknown modifications", {
    expect_warning(
        result <- convertAnnotation("M[UnknownMod]PEPTIDE",
                                   convertToStyle = "deltaMass"),
        "Could not find Unimod entry"
    )
    # Should return original when conversion fails
    expect_equal(result, "M[UnknownMod]PEPTIDE")
})


# Test internal helper functions ====

test_that(".detectModificationType correctly identifies modification types", {
    # Unimod ID formats
    expect_equal(PTMods:::.detectModificationType("UNIMOD:35"), "unimodId")
    expect_equal(PTMods:::.detectModificationType("U:35"), "unimodId")
    expect_equal(PTMods:::.detectModificationType("unimod:35"), "unimodId")

    # Delta mass formats
    expect_equal(PTMods:::.detectModificationType("+15.995"), "deltaMass")
    expect_equal(PTMods:::.detectModificationType("-17.026"), "deltaMass")
    expect_equal(PTMods:::.detectModificationType("+79.966331"), "deltaMass")
    expect_equal(PTMods:::.detectModificationType("+15"), "deltaMass")

    # Name format
    expect_equal(PTMods:::.detectModificationType("Oxidation"), "name")
    expect_equal(PTMods:::.detectModificationType("Phospho"), "name")
    expect_equal(PTMods:::.detectModificationType("Carbamidomethyl"), "name")
})

test_that(".extractUnimodId extracts numeric ID", {
    expect_equal(PTMods:::.extractUnimodId("UNIMOD:35"), 35)
    expect_equal(PTMods:::.extractUnimodId("U:35"), 35)
    expect_equal(PTMods:::.extractUnimodId("unimod:35"), 35)
    expect_equal(PTMods:::.extractUnimodId("UNIMOD:4"), 4)
})

test_that(".formatMass formats mass values correctly", {
    # Positive mass
    expect_equal(PTMods:::.formatMass(15.994915), "+15.994915")
    expect_equal(PTMods:::.formatMass(79.966331), "+79.966331")

    # Negative mass
    expect_equal(PTMods:::.formatMass(-17.026549), "-17.026549")

    # Zero
    expect_equal(PTMods:::.formatMass(0), "+0")

    # Custom precision
    expect_equal(PTMods:::.formatMass(15.994915, digits = 3), "+15.995")
    expect_equal(PTMods:::.formatMass(15.994915, digits = 2), "+15.99")
})

test_that(".lookupByName finds modifications correctly", {
    unimod_data <- modifications[!modifications$NeutralLoss,
                                 c("UnimodId", "Name", "MonoMass")]
    unimod_data <- unimod_data[!duplicated(unimod_data$Name), ]

    # Valid lookup
    result <- PTMods:::.lookupByName("Oxidation", unimod_data)
    expect_type(result, "list")
    expect_equal(result$unimodId, 35)
    expect_equal(result$name, "Oxidation")
    expect_equal(result$mass, 15.994915, tolerance = 0.000001)

    # Invalid lookup
    result <- PTMods:::.lookupByName("NonExistentMod", unimod_data)
    expect_null(result)
})

test_that(".lookupByMass finds modifications correctly", {
    unimod_data <- modifications[!modifications$NeutralLoss,
                                 c("UnimodId", "Name", "MonoMass")]
    unimod_data <- unimod_data[!duplicated(unimod_data$Name), ]

    # Valid lookup with tolerance
    result <- PTMods:::.lookupByMass(15.995, unimod_data, 0.01)
    expect_type(result, "list")
    expect_equal(result$unimodId, 35)
    expect_equal(result$name, "Oxidation")

    # Exact match
    result <- PTMods:::.lookupByMass(15.994915, unimod_data, 0.01)
    expect_type(result, "list")
    expect_equal(result$unimodId, 35)

    # No match with strict tolerance
    result <- PTMods:::.lookupByMass(999.999, unimod_data, 0.01)
    expect_null(result)
})

test_that(".lookupByUnimodId finds modifications correctly", {
    unimod_data <- modifications[!modifications$NeutralLoss,
                                 c("UnimodId", "Name", "MonoMass")]
    unimod_data <- unimod_data[!duplicated(unimod_data$Name), ]

    # Valid lookup
    result <- PTMods:::.lookupByUnimodId(35, unimod_data)
    expect_type(result, "list")
    expect_equal(result$unimodId, 35)
    expect_equal(result$name, "Oxidation")
    expect_equal(result$mass, 15.994915, tolerance = 0.000001)

    # Another valid lookup
    result <- PTMods:::.lookupByUnimodId(21, unimod_data)
    expect_type(result, "list")
    expect_equal(result$unimodId, 21)
    expect_equal(result$name, "Phospho")

    # Invalid lookup
    result <- PTMods:::.lookupByUnimodId(99999, unimod_data)
    expect_null(result)
})

test_that(".convertAnnotation handles single sequences", {
    # Valid conversion
    expect_equal(PTMods:::.convertAnnotation("M[Oxidation]PEPTIDE",
                                            convertToStyle = "deltaMass"),
                 "M[+15.994915]PEPTIDE")

    # Input validation
    expect_error(PTMods:::.convertAnnotation(c("A", "B"),
                                            convertToStyle = "deltaMass"),
                 "must be a single character string")

    expect_error(PTMods:::.convertAnnotation(123, convertToStyle = "deltaMass"),
                 "must be a single character string")
})

#' Managing amino acid modifications for mass spectrometry in R.
#'
#' The `PTMods` package focuses on handling post-translational modifications
#' (PTMs). It distributes PTMs from the Unimod database and allows to convert
#' PTM annotations between multiple formats.
#'
#' @name PTMods-package
#' @references \url{https://github.com/RforMassSpectrometry/PTMods/}
#' @keywords package
#'
#' @importFrom utils packageVersion
"_PACKAGE"

utils::globalVariables("aminoacids")


#' Calculate mass for a given aminoacid sequence
#'
#' @param x character, sequence
#' @param type character, mono or average mass
#' @return list of numeric (mass values)
#' @noRd
.aamass <- function(x, type=c("MonoMass", "AvgMass")) {
    m <- aminoacids[, match.arg(type)]
    names(m) <- aminoacids$OneLetter
    vapply(.string2character(x), function(xx)sum(m[xx]), NA_real_)
}

#' Calculate a single modification for a given sequence
#'
#' Throws a message if the default rule is applied and not verified (added to
#' the switch statement) by a human
#'
#' @param x character, sequence
#' @param id character, modification id
#' @param type character, mono or average mass
#' @param msg logical, show message if default was used?
#' @return numeric (mass values)
#' @noRd
.unimodMass <- function(x, id, type=c("MonoMass", "AvgMass"), msg=TRUE) {
    stopifnot(is.character(x) && is.character(id) && length(id) == 1L)
    stopifnot(id %in% modifications$Id)
    type <- match.arg(type)
    switch(id,
        ## 765: Met-loss
        "Met-loss:P-M" =
            # N-terminal initiator methionine is removed by a methionine
            # aminopeptidase from proteins where the residue following the
            # methionine is Ala, Cys, Gly, Pro, Ser, Thr or Val. This is
            # generally the final N-terminal state for proteins where the
            # following residue was a Cys, Pro or Val.
            grepl("^M([ACGPSTV])", x) * modifications[id, type],
        ## 766: Met-loss+Acetyl
        "Met-loss+Acetyl:P-M" =
            # The N-terminal initiator methionine is removed by a methionine
            # aminopeptidase from proteins whose residue following the
            # methionine is Ala, Cys, Gly, Pro, Ser, Thr or Val. Proteins
            # whose following residue was Ala, Gly, Ser or Thr are then
            # acetylated by an N(alpha)-acetyltransferase on the new
            # N-terminus.
            grepl("^M([AGST])", x) * modifications[id, type] +
            grepl("^M([CPV])", x) * modifications["Met-loss:P-M", type],
        ## default verified:
        ## 1: Acetyl
        "Acetyl:K" =, "Acetyl:N-term" =, "Acetyl:C" =, "Acetyl:S" =,
        "Acetyl:P-N-term" =, "Acetyl:T" =, "Acetyl:Y" =, "Acetyl:H" =,
        "Acetyl:R" =,
        ## 4: Carbamidomethyl
        "Carbamidomethyl:C" =, "Carbamidomethyl:K" =,
        "Carbamidomethyl:N-term" =, "Carbamidomethyl:H" =,
        "Carbamidomethyl:D" =, "Carbamidomethyl:E" =, "Carbamidomethyl:S" =,
        "Carbamidomethyl:T" =, "Carbamidomethyl:Y" =, "Carbamidomethyl:U" =,
        "Carbamidomethyl:M" =, "Carbamidomethyl:M:NL" =,
        ## 21: Phospho
        "Phospho:T" =, "Phospho:S" =, "Phospho:Y" =, "Phospho:D" =,
        "Phospho:H" =, "Phospho:C" =, "Phospho:R" =, "Phospho:K" =,
        "Phospho:T:NL" =, "Phospho:S:NL" =
        {
            .countSite(x, modifications[id, "Site"]) * modifications[id, type]
        },
        {
            if (msg) {
                message(
                    "Applying the default rule for the modification: ", id,
                    "\nPlease create an issue on: https://github.com/",
                    "ComputationalProteomicsUnit/unimod/issues/new ",
                    "\nto let us implement the correct rule or if the default ",
                    "one is already correct we could remove this message."
                )
            }
            .countSite(x, modifications[id, "Site"]) * modifications[id, type]
        }
    )
}

#' Modify the sequence for a single modification
#'
#' @param x character, sequence
#' @param id character, modification id
#' @return character (sequence)
#' @noRd
.unimodSequence <- function(x, id) {
    stopifnot(is.character(x) && is.character(id) && length(id) == 1L)
    stopifnot(id %in% modifications$Id)
    switch(id,
        ## 765: Met-loss
        "Met-loss:P-M" =,
        ## 766: Met-loss+Acetyl
        "Met-loss+Acetyl:P-M" =
            gsub("^M([ACGPSTV])", "\\1", x),
        x
    )
}

#' Find site occurrence
#'
#' @param x character, sequence
#' @param site character
#' @return numeric
#' @noRd
.countSite <- function(x, site) {
    stopifnot(is.character(x) && is.character(site) && length(site) == 1L)
    if (endsWith(site, "term")) {
        rep.int(1L, length(x))
    } else {
        nchar(x) - nchar(gsub(site, "", x, fixed=TRUE))
        ## while looking a little bit odd it is much faster than:
        # vapply(
        #     gregexpr(pattern=site, text=x, fixed=fixed),
        #     function(rx)sum(rx > 0L), double(1L),
        #     USE.NAMES=FALSE
        # )
    }
}

#' Calculate mass for a given aminoacid sequence and apply modifications
#'
#' @param x character, sequence
#' @param type character, mono or average mass
#' @param fixedModifications character/data.frame
#' @param variableModifications data.frame
#' @return list of numeric (mass values)
#' @noRd
.mass <- function(x, type=c("MonoMass", "AvgMass"),
                  fixedModifications=NULL,
                  variableModifications=NULL) {
    type <- match.arg(type)

    if (is.character(fixedModifications)) {
        i <- match(fixedModifications, modifications$Id)
        if (anyNA(i)) {
            stop(
                paste0(fixedModifications[is.na(i)], collapse=", "),
                ifelse(sum(is.na(i)) == 1L, " is ", " are "),
                "not part of the PTMods::modifications data set!"
            )
        }
        isUnimod <- TRUE
        fixedModifications <- modifications[i,,drop=FALSE]
    } else if (is.data.frame(fixedModifications)) {
        if (!all(c("Id", type, "Site") %in% colnames(fixedModifications))) {
            stop("The 'fixedModifications' `data.frame` has to have at least ",
                 "the columns 'Id', '", type, "', and 'Site'!")
        }
        isUnimod <- all(fixedModifications$Id %in% modifications$Id)
    } else if (!is.null(fixedModifications)) {
        stop("'fixedModifications' must be a `character`, `data.frame`, or ",
             "`NULL`!")
    }

    if (!is.null(variableModifications)) {
        stop("Variable Modifications are not supported yet!")
    }

    m <- .aamass(x, type)

    if (anyDuplicated(fixedModifications$Site)) {
        stop("Duplicated fixed modification sites are not allowed!")
    }

    if (!is.null(fixedModifications)) {
        for (i in seq_len(nrow(fixedModifications))) {
            if (isUnimod) {
                m <- m + .unimodMass(x, fixedModifications$Id[i], type=type)
                x <- .unimodSequence(x, fixedModifications$Id[i])
            } else {
                m <- m + .countSite(x, fixedModifications$Site[i]) *
                    fixedModifications[i, type]
            }
        }
    }
    attr(m, "sequence") <- x
    m
}

test_that(".aamass", {
    expect_error(PTMods:::.aamass(1:3))
    expect_error(PTMods:::.aamass("ACE", "FOO"))
    expect_equal(PTMods:::.aamass("ACE"), PTMods:::.aamass("ACE", "MonoMass"))
    expect_equal(PTMods:::.aamass("ACE"), 303.088892)
    expect_equal(PTMods:::.aamass("ACE", "AvgMass"), 303.3348)
    expect_equal(PTMods:::.aamass(c(foo="ACE", bar="CEA")),
                 c(foo=303.088892, bar=303.088892))
})

test_that(".unimodMass", {
    expect_error(PTMods:::.unimodMass(1:3))
    expect_error(PTMods:::.unimodMass("ACE"))
    expect_error(PTMods:::.unimodMass("ACE", "FOO"))
    expect_error(PTMods:::.unimodMass("ACE", c("FOO", "BAR")))
    expect_equal(PTMods:::.unimodMass("ACE", "Met-loss:P-M"), 0)
    expect_equal(PTMods:::.unimodMass("MACE", "Met-loss:P-M"),
                 PTMods:::.unimodMass("MACE", "Met-loss:P-M", "MonoMass"))
    expect_equal(PTMods:::.unimodMass("MACE", "Met-loss:P-M", "MonoMass"),
                 -131.040485)
    expect_equal(PTMods:::.unimodMass("MACE", "Met-loss:P-M", "AvgMass"),
                 -131.1961)
    expect_equal(PTMods:::.unimodMass("MCCE", "Met-loss:P-M"),
                 PTMods:::.unimodMass("MCCE", "Met-loss+Acetyl:P-M"))
    expect_equal(PTMods:::.unimodMass("MACE", "Met-loss+Acetyl:P-M"),
                 -89.02992)
    expect_equal(PTMods:::.unimodMass("ACE", "Acetyl:N-term"), 42.010565)
    expect_equal(PTMods:::.unimodMass("ACE", "Acetyl:C"), 42.010565)
    expect_equal(PTMods:::.unimodMass("ACE", "Acetyl:H"), 0)
    expect_equal(PTMods:::.unimodMass(c("ABE", "ACE"), "Acetyl:C"),
                 c(0, 42.010565))
    expect_message(PTMods:::.unimodMass("ACE", "Unknown:420:N-term"),
                   "Applying the default rule for .* create an issue")
    expect_silent(PTMods:::.unimodMass("ACE", "Unknown:420:N-term", msg=FALSE))
})

test_that(".unimodSequence", {
    expect_error(PTMods:::.unimodSequence(1:3))
    expect_error(PTMods:::.unimodSequence("ACE"))
    expect_error(PTMods:::.unimodSequence("ACE", "FOO"))
    expect_error(PTMods:::.unimodSequence("ACE", c("FOO", "BAR")))
    expect_equal(PTMods:::.unimodSequence("ACE", "Met-loss:P-M"), "ACE")
    expect_equal(PTMods:::.unimodSequence("MACE", "Met-loss:P-M"), "ACE")
    expect_equal(PTMods:::.unimodSequence("MACE", "Met-loss+Acetyl:P-M"),
                 "ACE")
    expect_equal(PTMods:::.unimodSequence(c("MBCE", "MACE"), "Met-loss:P-M"),
                 c("MBCE", "ACE"))
})

test_that(".mass", {
    expect_error(PTMods:::.mass(1:3))
    expect_error(PTMods:::.mass("ACE", "FOO"))
    expect_error(PTMods:::.mass("ACE", fixedModifications="FOO"), "is not part")
    expect_error(PTMods:::.mass("ACE",
                                fixedModifications=c("FOO", "Met-loss:P-M",
                                                     "BAR")), "are not part")
    expect_error(PTMods:::.mass("ACE", fixedModifications=1:3),
                 "must be a `character`")
    expect_error(PTMods:::.mass("ACE",
                                fixedModifications=c("Acetyl:K",
                                                     "Carbamidomethyl:K"),
                       "Duplicated fixed modification sites are not allowed"))
    r <- 303.088892
    attr(r, "sequence") <- "ACE"
    expect_equal(PTMods:::.mass("ACE"), r)
    expect_equal(PTMods:::.mass("ACE", fixedModifications="Acetyl:K"), r)
    r <- 345.099457
    attr(r, "sequence") <- "ACE"
    expect_equal(PTMods:::.mass("ACE", fixedModifications="Acetyl:N-term"), r)
    expect_equal(PTMods:::.mass("MACE",
                                fixedModifications="Met-loss+Acetyl:P-M"), r)
    r <- 402.120921
    attr(r, "sequence") <- "ACE"
    expect_equal(PTMods:::.mass("ACE",
                                fixedModifications=c("Acetyl:N-term",
                                                     "Carbamidomethyl:C")), r)
    r <- 478.119206
    attr(r, "sequence") <- "MDCE"
    expect_equal(PTMods:::.mass("MDCE", fixedModifications="Met-loss+Acetyl:P-M"), r)
})

test_that(".countSite", {
    expect_error(PTMods:::.countSite(1:3, "C"))
    expect_error(PTMods:::.countSite("ACE", 1:3))
    expect_error(PTMods:::.countSite("ACE", c("A", "C")))
    expect_equal(PTMods:::.countSite("ACE", "Nterm"), 1)
    expect_equal(PTMods:::.countSite("ACE", "Cterm"), 1)
    expect_equal(PTMods:::.countSite("ACE", "C"), 1)
    expect_equal(PTMods:::.countSite(c("ACCE", "ACE"), "Nterm"), c(1, 1))
    expect_equal(PTMods:::.countSite(c("ACCE", "ACE"), "Cterm"), c(1, 1))
    expect_equal(PTMods:::.countSite(c("ACCE", "ACE"), "C"), 2:1)
    expect_equal(PTMods:::.countSite(c("ACCE", "ACE"), "H"), c(0, 0))
})

#' Parse <elem> tag.
#' @param xml xml_document (returned by .unimodDb)
#' @return data.frame
#' @noRd
.elements <- function(xml) {
    stopifnot(requireNamespace("xml2"))
    elem <- xml2::xml_find_all(xml, "//umod:elements/umod:elem")
    m <- do.call(rbind, xml2::xml_attrs(elem))
    data.frame(
        Name=m[, "title"],
        FullName=m[, "full_name"],
        AvgMass=as.double(m[, "avge_mass"]),
        MonoMass=as.double(m[, "mono_mass"]),
        row.names=m[, "title"],
        stringsAsFactors=FALSE
    )
}

## TODO: save data.frame elements as public available data set `data(elements)`

test_that(".elements", {
    skip_if_not_installed("xml2")
    x <- xml2::read_xml(paste0(
        '<umod:unimod xmlns:umod="http://www.unimod.org/xmlns/schema/unimod_2">',
        '<umod:elements>',
        '<umod:elem title="H" full_name="Hydrogen" avge_mass="1.00794" mono_mass="1.007825035"/>',
        '<umod:elem title="2H" full_name="Deuterium" avge_mass="2.014101779" mono_mass="2.014101779"/>',
        '</umod:elements>',
        '</umod:unimod>'))
    d <- data.frame(
        Name=c("H", "2H"),
        FullName=c("Hydrogen", "Deuterium"),
        AvgMass=c(1.00794, 2.014101779),
        MonoMass=c(1.007825035, 2.014101779),
        row.names=c("H", "2H"),
        stringsAsFactors=FALSE
    )
    expect_equal(PTMods:::.elements(x), d)
})

#' @title Add fixed modifications
#'
#' @description Applies fixed modifications to peptide sequences.
#'
#' The annotation style of modifications already present in `sequences` is
#' preserved as-is.  The annotation style of newly applied modifications
#' matches the style supplied in `fixedModifications`: numeric values produce
#' deltaMass notation (e.g. `[+79.966331]`), while character values are used
#' verbatim (e.g. `[Phospho]` or `[UNIMOD:21]`).
#'
#' @param sequences Character vector. Peptide sequences that may have
#' modifications or not.
#'
#' @param fixedModifications Named `numeric` or `character`. If a
#' `character` is given, values must be in UniMod name or UniMod ID format
#' (e.g. `"Phospho"`, `"UNIMOD:21"`). The annotation style of the values is
#' preserved in the output.
#' Specifies which fixed modifications are applied to which amino acids. By
#' default, carbamidomethylation is applied (+57.021464 on cysteines).
#' Supplying multiple modifications for the same amino acid is not supported;
#' call the function consecutively instead:
#' \preformatted{
#'   "ATK" |>
#'     addFixedModifications(c(T = "+57.024")) |>
#'     addFixedModifications(c(T = "Phospho"))
#' }
#'
#' @return A named list of character strings with all fixed modifications
#' applied to the corresponding amino acids, one element per input sequence.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @export
#'
#' @examples
#' addFixedModifications("ATCK")
#' addFixedModifications("ATCK", fixedModifications = c(Nterm = 304, C = 57.02146))
#' addFixedModifications("ATSK", fixedModifications = c(A = -42, S = 79.966331))
#' addFixedModifications("ATK", fixedModifications = c(T = "Phospho", A = 42))
#' addFixedModifications("A[+42]TK", fixedModifications = c(T = "Phospho"))
#' addFixedModifications("AT[+79.966]AK", fixedModifications = c(A = 42))
#' addFixedModifications("ATA[+79.966]K", fixedModifications = c(A = 42))
#'
#' ## Apply two modifications to the same residue consecutively
#' "ATK" |>
#'     addFixedModifications(c(T = "+57.024")) |>
#'     addFixedModifications(c(T = "Phospho"))
addFixedModifications <- function(sequences,
                                  fixedModifications = c(C = 57.021464)) {
    vapply(sequences, .addFixedModifications,
        FUN.VALUE = character(1L),
        fixedModifications = fixedModifications,
        USE.NAMES = FALSE
    )
}

#' @title Applies fixed modifications to a peptide sequence
#'
#' @param sequence Character. A peptide sequence that may have modifications or
#' not.
#'
#' @param fixedModifications Named `numeric` or `character`. If a
#' `character` is given, it needs to be in unimodId or name format as
#' specified in `convertAnnotation`.
#' Specifies which fixed modifications are applied to which amino acids.
#' Use Nterm or Cterm as names for modifications that should be added to the
#' amino/carboxyl-termini. Default is set to carbamidomethylation
#' (C = 57.02146). Supplying multiple modifications for the same amino acid
#' is not supported; call `addFixedModifications()` consecutively instead:
#' \preformatted{
#'   "ATK" |>
#'     addFixedModifications(c(T = "+57.024")) |>
#'     addFixedModifications(c(T = "Phospho"))
#' }
#'
#' @return A character string with all fixed modifications applied to the
#' corresponding amino acids.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @importFrom stats setNames
#'
#' @noRd
#'
#' @examples
#' .addFixedModifications("ATCK", fixedModifications = c(Nterm = 304, C = 57.02146))
#' .addFixedModifications("ATSK", fixedModifications = c(A = -42, S = 79.966331))
#' .addFixedModifications("ATK", fixedModifications = c(T = "Phospho", A = "UNIMOD:1"))
#' .addFixedModifications("AT[+79.966]AK", fixedModifications = c(A = 42))
#' .addFixedModifications("ATA[+79.966]K", fixedModifications = c(A = 42))
#'
#' ## Apply two modifications to the same residue consecutively
#' "ATK" |>
#'     addFixedModifications(c(T = "+57.024")) |>
#'     addFixedModifications(c(T = "Phospho"))
.addFixedModifications <- function(sequence,
                                   fixedModifications = c(C = 57.02146)) {
    if (is.na(sequence)) {
        return(NA_character_)
    }

    if (!is.null(fixedModifications)) {
        mod_table <- table(names(fixedModifications))

        if (!all(mod_table == 1)) {
            stop(
                "Applying multiple modifications to the same amino acid is not ",
                "supported.\n",
                "Please use addFixedModifications() consecutively instead:\n\n",
                "  'ATK' |>\n",
                "    addFixedModifications(c(T = '+57.024')) |>\n",
                "    addFixedModifications(c(T = 'Phospho'))"
            )
        }
    }

    terms <- .extractTerminalMods(sequence)
    sequence <- terms$sequence

    sequence <- regmatches(
        sequence,
        gregexpr("[A-Z](\\[[^\\]]*\\])*", sequence, perl = TRUE)
    )[[1]]
    seqAA <- sub("\\[.*", "", sequence)
    modifiedAA <- names(fixedModifications)

    fixedMods <- .resolveModsToStrings(fixedModifications)
    names(fixedMods) <- modifiedAA

    mStr <- setNames(rep(NA_character_, length(sequence)), sequence)
    applicable <- which(seqAA %in% modifiedAA)
    if (length(applicable)) {
        mStr[applicable] <- fixedMods[seqAA[applicable]]
    }

    sequence <- paste(
        ifelse(!is.na(mStr),
            paste0(names(mStr), "[", mStr, "]"),
            names(mStr)
        ),
        collapse = ""
    )

    sequence <- .addTerminalModifications(sequence, fixedMods)
    .reattachTerminalMods(sequence, terms$nterm, terms$cterm)
}

#' Adds nterm/cterm modifications to sequence
#'
#' @param sequence A `character(1L)` with a peptide sequence with or without
#' modifications.
#'
#' @param modifications A named `numeric` or `character` vector with all
#' modifications to be applied on the peptide sequence. Numeric values are
#' formatted as deltaMass (e.g. `[+67]`); character values are used verbatim
#' (e.g. `[Phospho]` or `[UNIMOD:21]`).
#'
#' @return The same sequence with additional Nterm or Cterm modifications if
#' present in `modifications`.
#'
#' @noRd
#'
#' @examples
#' PTMods:::.addTerminalModifications("ATK", c(Nterm = 67))
#' PTMods:::.addTerminalModifications("ATK", c(Nterm = "Acetyl"))
.addTerminalModifications <- function(sequence, modifications) {
    modStrings <- .resolveModsToStrings(modifications)

    if ("Nterm" %in% names(modStrings)) {
        val <- modStrings[["Nterm"]]
        sequence <- paste0("[", val, "]-", sequence)
    }

    if ("Cterm" %in% names(modStrings)) {
        val <- modStrings[["Cterm"]]
        sequence <- paste0(sequence, "-[", val, "]")
    }

    sequence
}

# Test addFixedModifications

test_that("addFixedModifications applies numeric fixed modification", {
    result <- addFixedModifications("ATCK", fixedModifications = c(C = 57.02146))
    expect_equal(result, "ATC[+57.02146]K")
})

test_that("addFixedModifications applies multiple numeric modifications", {
    result <- addFixedModifications("ATSK",
        fixedModifications = c(
            A = -42,
            S = 79.966331
        )
    )
    expect_equal(result, "A[-42]TS[+79.966331]K")
})

test_that("addFixedModifications applies character name modifications preserving style", {
    result <- addFixedModifications("ATK",
        fixedModifications = c(
            T = "Phospho",
            A = "UNIMOD:1"
        )
    )
    ## Annotation style of the parameter is preserved in the output
    expect_equal(result, "A[UNIMOD:1]T[Phospho]K")
})

test_that("addFixedModifications preserves existing annotation style while adding character mods", {
    result <- addFixedModifications("A[+42]TK",
        fixedModifications = c(T = "Phospho")
    )
    ## Existing [+42] stays in deltaMass, new mod uses name style as supplied
    expect_equal(result, "A[+42]T[Phospho]K")
})

test_that("addFixedModifications preserves existing modifications", {
    result <- addFixedModifications("AT[+79.966]AK",
        fixedModifications = c(A = 42)
    )
    expect_equal(result, "A[+42]T[+79.966]A[+42]K")
})

test_that("addFixedModifications stacks on already-modified residues", {
    result <- addFixedModifications("ATA[+79.966]K",
        fixedModifications = c(A = 42)
    )
    expect_equal(result, "A[+42]TA[+79.966][+42]K")
})

test_that("addFixedModifications handles vectorised input", {
    result <- addFixedModifications(c("ATCK", "PEPCK"),
        fixedModifications = c(C = 57.02146)
    )
    expect_length(result, 2)
    expect_equal(result, c("ATC[+57.02146]K", "PEPC[+57.02146]K"))
    expect_null(names(result))
})

test_that("addFixedModifications handles sequences without target amino acids", {
    result <- addFixedModifications("ATGK", fixedModifications = c(C = 57.02146))
    expect_equal(result, "ATGK")
})


test_that("addFixedModifications adds + prefix for numeric-as-character values", {
    result <- addFixedModifications("ATCK",
        fixedModifications = c(Nterm = "Phospho", C = 57.02146)
    )
    expect_equal(result, "[Phospho]-ATC[+57.02146]K")
})

test_that("addFixedModifications preserves - prefix for negative numeric-as-character values", {
    result <- addFixedModifications("ATK",
        fixedModifications = c(A = "-42", T = "79.966331")
    )
    expect_equal(result, "A[-42]T[+79.966331]K")
})

test_that("addFixedModifications applies two mods to same residue consecutively", {
    result <- "ATK" |>
        addFixedModifications(c(T = "+57.024")) |>
        addFixedModifications(c(T = "Phospho"))
    expect_equal(result, "AT[+57.024][Phospho]K")
})

# Test internal .addFixedModifications

test_that(".addFixedModifications applies N-terminal modification", {
    result <- PTMods:::.addFixedModifications("ATCK",
        fixedModifications = c(Nterm = 304, C = 57.02146)
    )
    expect_equal(result, "[+304]-ATC[+57.02146]K")
})

test_that(".addFixedModifications applies C-terminal modification", {
    result <- PTMods:::.addFixedModifications("ATCK",
        fixedModifications = c(Cterm = 42)
    )
    expect_equal(result, "ATCK-[+42]")
})

test_that(".addFixedModifications applies both terminal modifications", {
    result <- PTMods:::.addFixedModifications("ATK",
        fixedModifications = c(Nterm = 304, Cterm = 42)
    )
    expect_equal(result, "[+304]-ATK-[+42]")
})

test_that(".addFixedModifications preserves existing terminal modifications", {
    result <- PTMods:::.addFixedModifications("[+10]-ATK",
        fixedModifications = c(A = 42)
    )
    expect_equal(result, "[+10]-A[+42]TK")
})

test_that(".addFixedModifications handles negative modifications", {
    result <- PTMods:::.addFixedModifications("ATK",
        fixedModifications = c(A = -42)
    )
    expect_equal(result, "A[-42]TK")
})


# Test internal .addTerminalModifications

test_that(".addTerminalModifications adds N-terminal modification", {
    result <- PTMods:::.addTerminalModifications("ATK", c(Nterm = 67))
    expect_equal(result, "[+67]-ATK")
})

test_that(".addTerminalModifications adds C-terminal modification", {
    result <- PTMods:::.addTerminalModifications("ATK", c(Cterm = 42))
    expect_equal(result, "ATK-[+42]")
})

test_that(".addTerminalModifications adds both terminal modifications", {
    result <- PTMods:::.addTerminalModifications(
        "ATK",
        c(Nterm = 304, Cterm = 42)
    )
    expect_equal(result, "[+304]-ATK-[+42]")
})

test_that(".addTerminalModifications skips if no terminal mods", {
    result <- PTMods:::.addTerminalModifications("ATK", c(A = 42))
    expect_equal(result, "ATK")
})

test_that("addFixedModifications preserves NAs in input", {
    result <- addFixedModifications(c(NA, "ATK"))
    expect_true(is.na(result[[1]]))
    expect_equal(result[[2]], "ATK")
})

#' @title Get canonical sequence from any peptidoform
#'
#' @description Remove any modifications annotation on a sequence. Annotation
#' from any type (so long as they are specified within brackets "[..]") are
#' removed.
#'
#' @param x `character`, ProForma sequence.
#'
#' @return `character`, a `character` vector with sequences cleaned of all
#' modifications.
#'
#' @export
#'
#' @examples
#' getCanonicalSequence(
#'     c(
#'         "EM[-15.9949]EVEES[+79.9663]PEK",
#'         "EK[Acetyl]EVEES[UNIMOD:21]PEK"
#'     )
#' )
getCanonicalSequence <- function(x) {
    x <- gsub("\\[.*?\\]", "", x)
    gsub("^-|-$", "", x) ## Remove any hyphens due to terminal mods
}

#' @title Extract modification counts from one or more peptide sequences
#'
#' @description
#' Given a character vector of peptide sequences containing modifications in
#' any annotation style (deltaMass, UniMod ID, or name), returns a named
#' integer vector with the count of each unique modification. Regardless of
#' the input annotation style, all modifications are converted to their
#' UniMod name in the output (e.g. `[+79.966]` and `[UNIMOD:21]` both
#' appear as `Phospho`).
#'
#' @param sequences A `character()` vector of peptide sequences in ProForma
#' format.
#'
#' @return A named integer vector where names are UniMod modification names
#' and values are their occurrence counts across all input sequences.
#' Returns an empty integer vector if no modifications are found.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @examples
#' seqs <- c(
#'     "[+304]-AT[Phospho]K",
#'     "AC[Carbamidomethyl]T[+79.966]S[Phospho]K",
#'     "PEPTIDE"
#' )
#' getModificationsCounts(seqs)
#'
#' @export
getModificationsCounts <- function(sequences) {
    matches <- .getModifications(sequences)

    if (!length(matches)) {
        return(setNames(integer(0L), character(0L)))
    }

    matches <- convertAnnotation(matches, "name", verbose = FALSE)

    mods <- gsub("^-?\\[|\\]$", "", matches)
    table_counts <- table(mods)
    setNames(as.integer(table_counts), names(table_counts))
}

#' Extract raw modification strings from peptide sequences
#'
#' Collapses all input sequences into one string and extracts every
#' bracket-enclosed modification token. The strings are returned in their
#' original annotation style (deltaMass, UniMod ID, or name) without any
#' conversion. Conversion to UniMod name occurs in the calling function
#' `getModificationsCounts`.
#'
#' @param sequences A `character()` vector of peptide sequences in ProForma
#'   format.
#'
#' @return A `character()` vector of raw modification tokens including
#'   their enclosing brackets (e.g. `"[+79.966]"`, `"[Phospho]"`,
#'   `"[UNIMOD:21]"`). Returns `character(0)` if no modifications are
#'   found.
#'
#' @noRd
.getModifications <- function(sequences) {
    stopifnot(is.character(sequences))

    collapsed <- paste(sequences, collapse = "")
    regmatches(
        collapsed,
        gregexpr("\\[([^\\[\\]]+)\\]", collapsed, perl = TRUE)
    )[[1L]]
}

#' @title Parse and strip terminal modifications from a sequence
#'
#' @param sequence `character(1L)`. A peptide sequence possibly containing
#' N-terminal (e.g. `[+304]-`) and/or C-terminal (e.g. `-[+42]`) modifications.
#'
#' @return A named `list` with elements:
#' \describe{
#'   \item{`sequence`}{The inner sequence with terminal mods removed.}
#'   \item{`nterm`}{`character(1L)` N-terminal prefix (e.g. `"[+304]-"`), or `""`.}
#'   \item{`cterm`}{`character(1L)` C-terminal suffix (e.g. `"-[+42]"`), or `""`.}
#' }
#'
#' @noRd
#'
#' @examples
#' PTMods:::.extractTerminalMods("[+304]-ATK")
#' PTMods:::.extractTerminalMods("ATK-[+42]")
#' PTMods:::.extractTerminalMods("[+304]-ATK-[+42]")
#' PTMods:::.extractTerminalMods("ATK")
.extractTerminalMods <- function(sequence) {
    nterm <- ""
    cterm <- ""

    ## N-terminal: optional bracket expression followed by "-"
    nterm_match <- regmatches(sequence, regexpr("^(\\[[^\\]]*\\])-",
        sequence,
        perl = TRUE
    ))
    if (length(nterm_match) && nchar(nterm_match) > 0L) {
        nterm <- nterm_match
        sequence <- substring(sequence, nchar(nterm_match) + 1L)
    }

    ## C-terminal: "-" followed by optional bracket expression at end
    cterm_match <- regmatches(sequence, regexpr("-(\\[[^\\]]*\\])$",
        sequence,
        perl = TRUE
    ))
    if (length(cterm_match) && nchar(cterm_match) > 0L) {
        cterm <- cterm_match
        sequence <- substring(
            sequence, 1L,
            nchar(sequence) - nchar(cterm_match)
        )
    }

    list(sequence = sequence, nterm = nterm, cterm = cterm)
}

#' @title Reattach terminal modifications to a sequence
#'
#' @param sequence `character(1L)`. Inner peptide sequence without terminal mods.
#' @param nterm `character(1L)`. N-terminal prefix to prepend (e.g. `"[+304]-"`).
#' @param cterm `character(1L)`. C-terminal suffix to append (e.g. `"-[+42]"`).
#'
#' @return `character(1L)` with terminal mods reattached.
#'
#' @noRd
#'
#' @examples
#' PTMods:::.reattachTerminalMods("ATK", "[+304]-", "")
#' PTMods:::.reattachTerminalMods("ATK", "", "-[+42]")
.reattachTerminalMods <- function(sequence, nterm, cterm) {
    paste0(nterm, sequence, cterm)
}

#' @title Convert modifications to bracket-content strings, preserving style
#'
#' @description Converts a named `numeric` or `character` modifications vector
#' to a named `character` vector of bracket contents:
#' \itemize{
#'   \item `numeric` values are formatted as `"+N"` or `"-N"` (deltaMass style).
#'   \item `character` values are returned unchanged, preserving `name` or
#'   `unimodId` annotation style.
#' }
#'
#' @param modifications Named `numeric` or `character`. Modifications to
#' resolve.
#'
#' @return Named `character` vector of bracket-content strings.
#'
#' @noRd
#'
#' @examples
#' PTMods:::.resolveModsToStrings(c(A = 42.010565, T = -15.994915))
#' PTMods:::.resolveModsToStrings(c(T = "Phospho", A = "UNIMOD:1"))
.resolveModsToStrings <- function(modifications) {
    if (is.numeric(modifications)) {
        result <- vapply(modifications, function(m) {
            if (m >= 0) paste0("+", m) else as.character(m)
        }, character(1L))
        names(result) <- names(modifications)
        result
    } else {
        ## character vector: elements that look like numbers (e.g. "57.02146")
        ## get the same +/- prefix treatment as numeric values; other elements
        ## (e.g. "Phospho", "UNIMOD:21") are returned verbatim.
        result <- vapply(modifications, function(m) {
            num <- suppressWarnings(as.numeric(m))
            if (!is.na(num) && num >= 0 && !startsWith(m, "+")) {
                paste0("+", m)
            } else {
                m
            }
        }, character(1L))
        names(result) <- names(modifications)
        result
    }
}

## Test getCanonicalSequence

test_that("getCanonicalSequence removes delta mass modifications", {
    expect_equal(
        getCanonicalSequence("EM[-15.9949]EVEES[+79.9663]PEK"),
        "EMEVEESPEK"
    )
    expect_equal(
        getCanonicalSequence("APEPT[+79.966]IDEK"),
        "APEPTIDEK"
    )
})

test_that("getCanonicalSequence removes name modifications", {
    expect_equal(
        getCanonicalSequence("EK[Acetyl]EVEES[Phospho]PEK"),
        "EKEVEESPEK"
    )
    expect_equal(
        getCanonicalSequence("M[Oxidation]PEPTIDE"),
        "MPEPTIDE"
    )
})

test_that("getCanonicalSequence removes unimod ID modifications", {
    expect_equal(
        getCanonicalSequence("EK[UNIMOD:1]EVEES[UNIMOD:21]PEK"),
        "EKEVEESPEK"
    )
})

test_that("getCanonicalSequence removes terminal modifications", {
    expect_equal(getCanonicalSequence("[+304]-ATCK"), "ATCK")
    expect_equal(getCanonicalSequence("ATCK-[+42]"), "ATCK")
    expect_equal(getCanonicalSequence("[+304]-ATCK-[+42]"), "ATCK")
})

test_that("getCanonicalSequence handles unmodified sequences", {
    expect_equal(getCanonicalSequence("PEPTIDE"), "PEPTIDE")
    expect_equal(getCanonicalSequence("ATK"), "ATK")
})

test_that("getCanonicalSequence handles vectorised input", {
    result <- getCanonicalSequence(c(
        "EM[-15.9949]EVEES[+79.9663]PEK",
        "EK[Acetyl]EVEES[UNIMOD:21]PEK",
        "PEPTIDE"
    ))
    expect_equal(result, c("EMEVEESPEK", "EKEVEESPEK", "PEPTIDE"))
})

test_that("getCanonicalSequence removes stacked modifications", {
    expect_equal(getCanonicalSequence("AT[+10][-5]K"), "ATK")
    expect_equal(getCanonicalSequence("A[Phospho][UNIMOD:1]TK"), "ATK")
})


# Test internal helpers

test_that(".extractTerminalMods parses N-terminal modifications", {
    result <- PTMods:::.extractTerminalMods("[+304]-ATK")
    expect_equal(result$sequence, "ATK")
    expect_equal(result$nterm, "[+304]-")
    expect_equal(result$cterm, "")
})

test_that(".extractTerminalMods parses C-terminal modifications", {
    result <- PTMods:::.extractTerminalMods("ATK-[+42]")
    expect_equal(result$sequence, "ATK")
    expect_equal(result$nterm, "")
    expect_equal(result$cterm, "-[+42]")
})

test_that(".extractTerminalMods parses both terminal modifications", {
    result <- PTMods:::.extractTerminalMods("[+304]-ATK-[+42]")
    expect_equal(result$sequence, "ATK")
    expect_equal(result$nterm, "[+304]-")
    expect_equal(result$cterm, "-[+42]")
})

test_that(".extractTerminalMods handles no terminal modifications", {
    result <- PTMods:::.extractTerminalMods("ATK")
    expect_equal(result$sequence, "ATK")
    expect_equal(result$nterm, "")
    expect_equal(result$cterm, "")
})

test_that(".extractTerminalMods preserves internal modifications", {
    result <- PTMods:::.extractTerminalMods("[+304]-AT[+79.966]K-[+42]")
    expect_equal(result$sequence, "AT[+79.966]K")
    expect_equal(result$nterm, "[+304]-")
    expect_equal(result$cterm, "-[+42]")
})

test_that(".reattachTerminalMods prepends and appends correctly", {
    expect_equal(
        PTMods:::.reattachTerminalMods("ATK", "[+304]-", ""),
        "[+304]-ATK"
    )
    expect_equal(
        PTMods:::.reattachTerminalMods("ATK", "", "-[+42]"),
        "ATK-[+42]"
    )
    expect_equal(
        PTMods:::.reattachTerminalMods("ATK", "[+304]-", "-[+42]"),
        "[+304]-ATK-[+42]"
    )
    expect_equal(
        PTMods:::.reattachTerminalMods("ATK", "", ""),
        "ATK"
    )
})

#' @title Add fixed and variable modifications to peptide sequences
#'
#' @description
#' A convenience wrapper around [addFixedModifications()] and
#' [addVariableModifications()] that applies both fixed and variable
#' modifications in a single call. Fixed modifications are applied first,
#' followed by the enumeration of all possible combinations of variable
#' modifications. Optionally, the annotation style of
#' the output can be converted via [convertAnnotation()].
#'
#' @param sequences `character`. Peptide sequences with or without existing
#'   modification annotations. Modifications must be enclosed in square
#'   brackets (e.g. `"AT[+79.966]K"`, `"[+304]-ATK"`).
#'
#' @param fixedModifications Named `numeric` or `character`, or `NULL`.
#'   Fixed modifications to be applied unconditionally to the specified amino
#'   acids or termini. Names correspond to single-letter amino acid codes,
#'   `"Nterm"`, or `"Cterm"`. If `character`, values must be in UniMod name
#'   or UniMod ID format (e.g. `"Phospho"`, `"UNIMOD:21"`). See
#'   [addFixedModifications()] for full details. Set to `NULL` to skip.
#'
#' @param variableModifications Named `numeric` or `character`, or `NULL`.
#'   Variable modifications to enumerate over the specified amino acids.
#'   Names correspond to single-letter amino acid codes. If `character`,
#'   values must be in UniMod name or UniMod ID format. See
#'   [addVariableModifications()] for full details. Set to `NULL` to skip.
#'
#' @param convertToStyle `character(1)`. Controls the annotation format of
#'   modifications in the returned sequences. One of:
#'   \describe{
#'     \item{`NULL`}{(default) Same as input modification style.}
#'     \item{`"deltaMass"`}{Delta mass notation, e.g. `[+79.966]`.}
#'     \item{`"unimodId"`}{UniMod ID notation, e.g. `[UNIMOD:21]`.}
#'     \item{`"name"`}{UniMod name notation, e.g. `[Phospho]`.}
#'   }
#'   Conversion is performed by [convertAnnotation()].
#'
#' @param ... Additional arguments passed to the underlying modification
#'   functions. Currently supported:
#'   \describe{
#'     \item{`maxMods`}{Forwarded to [addVariableModifications()]. A
#'       `numeric(1)` specifying the maximum number of variable modifications
#'       that may be applied simultaneously. Defaults to `Inf`.}
#'   }
#'
#' @return A `character` vector of peptidoforms. The length is greater than or
#'   equal to `length(sequences)`, as each input sequence may expand into
#'   multiple peptidoforms corresponding to all combinations of variable
#'   modifications.
#'
#' @seealso
#' * \code{\link{addFixedModifications}} for applying fixed modifications.
#'
#' * \code{\link{addVariableModifications}} for enumerating variable modification
#'   combinations.
#'
#' * \code{\link{convertAnnotation}} for converting between modification annotation
#'   styles.
#'
#' @author Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
#'
#' @export
#'
#' @examples
#' ## Fixed modifications only
#' addModifications(
#'     c("ATCK", "PEPCK"),
#'     fixedModifications = c(C = 57.02146)
#' )
#'
#' ## Variable modifications only
#' addModifications(
#'     c("ATSK", "PEPTSK"),
#'     variableModifications = c(S = 79.966331, T = 79.966331),
#'     maxMods = 2
#' )
#'
#' ## Both fixed and variable modifications
#' addModifications(
#'     "ATCSK",
#'     fixedModifications = c(C = 57.02146),
#'     variableModifications = c(S = 79.966331, T = 79.966331),
#'     maxMods = 1
#' )
#'
#' ## Return annotations in UniMod name style
#' addModifications(
#'     "ATSK",
#'     variableModifications = c(S = 79.966331, T = 79.966331),
#'     convertToStyle = "name"
#' )
#'
#' ## N-terminal fixed modification with variable modifications
#' addModifications(
#'     "ATSK",
#'     fixedModifications = c(Nterm = 304),
#'     variableModifications = c(S = 79.966331)
#' )
addModifications <- function(sequences,
                             fixedModifications = NULL,
                             variableModifications = NULL,
                             convertToStyle = NULL,
                             ...) {
    dots <- list(...)

    ## --- Fixed modifications --- ##
    if (!is.null(fixedModifications)) {
        sequences <- addFixedModifications(
            sequences,
            fixedModifications = fixedModifications
        )
    }

    ## --- Variable modifications ---##
    if (!is.null(variableModifications)) {
        maxMods <- if (!is.null(dots[["maxMods"]])) dots[["maxMods"]] else Inf
        sequences <- addVariableModifications(
            sequences,
            variableModifications = variableModifications,
            maxMods = maxMods
        )
    }

    ## --- Convert annotation style --- ##
    if (!is.null(convertToStyle)) {
        sequences <- convertAnnotation(sequences,
            convertToStyle = convertToStyle
        )
    }

    sequences
}

# Test addModifications

test_that("addModifications with fixed modifications only", {
    result <- addModifications(
        c("ATCK", "PEPCK"),
        fixedModifications = c(C = 57.02146)
    )
    expect_equal(result, c("ATC[+57.02146]K", "PEPC[+57.02146]K"))
})

test_that("addModifications with variable modifications only", {
    result <- addModifications(
        "ATSK",
        variableModifications = c(S = 79.966331, T = 79.966331),
        maxMods = 1
    )
    ## Modifiable: T, S → 1+2 = 3 combos
    expect_length(result, 3)
    expect_true("ATSK" %in% result)
    expect_true("AT[+79.966331]SK" %in% result)
    expect_true("ATS[+79.966331]K" %in% result)
})

test_that("addModifications with both fixed and variable modifications", {
    result <- addModifications(
        "ATCSK",
        fixedModifications = c(C = 57.02146),
        variableModifications = c(S = 79.966331, T = 79.966331),
        maxMods = 1
    )
    ## Fixed first: ATC[+57.02146]SK
    ## Then variable on T and S: 1+2 = 3 combos
    expect_length(result, 3)
    expect_true("ATC[+57.02146]SK" %in% result)
    expect_true("AT[+79.966331]C[+57.02146]SK" %in% result)
    expect_true("ATC[+57.02146]S[+79.966331]K" %in% result)
})

test_that("addModifications returns annotations in name style", {
    result <- addModifications(
        "ATSK",
        variableModifications = c(S = 79.966331, T = 79.966331),
        convertToStyle = "name",
        maxMods = 1
    )
    expect_true("ATSK" %in% result)
    expect_true("AT[Phospho]SK" %in% result)
    expect_true("ATS[Phospho]K" %in% result)
})

test_that("addModifications returns annotations in unimodId style", {
    result <- addModifications(
        "ATSK",
        variableModifications = c(S = 79.966331, T = 79.966331),
        convertToStyle = "unimodId",
        maxMods = 1
    )
    expect_true("ATSK" %in% result)
    expect_true("AT[UNIMOD:21]SK" %in% result)
    expect_true("ATS[UNIMOD:21]K" %in% result)
})

test_that("addModifications with N-terminal fixed modification", {
    result <- addModifications(
        "ATSK",
        fixedModifications = c(Nterm = 304),
        variableModifications = c(S = 79.966331)
    )
    ## Fixed: [+304]-ATSK
    ## Variable on S: 2 combos
    expect_length(result, 2)
    expect_true("[+304]-ATSK" %in% result)
    expect_true("[+304]-ATS[+79.966331]K" %in% result)
})

test_that("addModifications with NULL fixed and variable returns unchanged", {
    result <- addModifications("PEPTIDE",
        fixedModifications = NULL,
        variableModifications = NULL
    )
    expect_equal(result, "PEPTIDE")
})

test_that("addModifications passes maxMods through dots", {
    result <- addModifications(
        "ATSK",
        variableModifications = c(A = 4, T = 5, S = 8),
        maxMods = 2
    )
    ## Modifiable:
    ## 0: 1, 1-mod: 3, 2-mod: 3 → total 7
    expect_length(result, 7)
})

---
title: "The PTMods package: a package to handle post-translational modifications"
output:
    BiocStyle::html_document:
        toc_float: true
vignette: >
    %\VignetteIndexEntry{The PTMods package}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
    %\VignettePackage{PTMods}
    %\VignetteDepends{PTMods}
---

```{r style, echo = FALSE, results = 'asis', message=FALSE}
BiocStyle::markdown()
```

**Package**: `r Biocpkg("PTMods")`<br />
**Authors**: `r packageDescription("PTMods")[["Author"]] `<br />
**Last modified:** `r file.info("PTMods.Rmd")$mtime`<br />
**Compiled**: `r date()`

# Introduction

Post-translational modifications (PTMs) are covalent modifications
that happen after protein biosynthesis and that can have a major
influence on protein function. PTMs are biologically very important
and can be identified in a high-throughput way using
mass spectrometry.

The `PTMods` package focuses on handling such PTMs and works in
conjuction with the `r BiocStyle::Biocpkg("PSMatch")` and
`r BiocStyle::Biocpkg("Spectra")` packages. It distributes PTMs from
the Unimod database and allows to convert PTM annotations between
multiple annotation formats. The package also allows the seemless addition of
fixed and/or variable PTMs using different annotation formats.

```{r lib}
library(PTMods)
```

```{r loaddata, echo = FALSE}
data(aminoacids)
data(modifications)
data(elements)
```
## Installation

The package can be installed from Bioconductor with

```{r install, eval = FALSE}
if (!require("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}

BiocManager::install("PTMods")
```

Alternatively, altough this isn't recommended in general, it can also
be installed from Github.

```{r install2o, eval = FALSE}
BiocManager::install("RforMassSpectrometry/PTMods")
```

# The Unimod database

The `PTMods` package distributes the data from the Unimod database:

> Unimod is a public domain database, distributed under a
> [copyleft licence](http://www.gnu.org/licenses/license-list.html): "a
> copyright notice that permits unrestricted redistribution and
> modification, provided that all copies and derivatives retain the same
> permissions."
>
> The aim is to create a community supported, comprehensive database of
> protein modifications for mass spectrometry applications. That is,
> accurate and verifiable values, derived from elemental compositions,
> for the mass differences introduced by all types of natural and
> artificial modifications. Other important information includes any
> mass change, (neutral loss), that occurs during MS/MS analysis, and
> site specificity, (which residues are susceptible to modification and
> any constraints on the position of the modification within the protein
> or peptide).

Source: http://www.unimod.org/unimod_help.html
We try to keep the data up to date with the Unimod database. If you believe it
isn't, do not hesitate to contact us through an 
[issue on the GitHub page](https://github.com/rformassspectrometry/PTMods/issues/new).

The package provides the following objects.

- The `modifications` dataframe describing `r nrow(modifications)`
  PTMs from the unimod database:

```{r modifications}
data(modifications)
head(modifications)
```

- The `aminoacids` dataframe describing `r nrow(aminoacids)` amino
  acids from the unimod database:

```{r aminoacids}
data(aminoacids)
head(aminoacids)
```
- The `elements` data.frame describing `r nrow(elements)` chemical
  elements from the unimod database:

```{r elements}
data(elements)
head(elements)
```

Note that the goal of `PTMods` is not to place this information in a
biological context. There are other sources of biological information
about post-translational modifications:


- [RESID](https://proteininformationresource.org/resid/resid.shtml)
  for post-translational modifications
- [Uniprot/Swiss-Prot](https://www.uniprot.org/) protein sequence
  database
- [Prosite](https://prosite.expasy.org/) database of protein families
  and domains
- [Glycan Database](https://www.functionalglycomics.org/) from the
  Consortium for Functional Glycomics
- [PhosphoDB](https://phosphodb.hecklab.com/site/index) and
  [PhosphoSitePlus](https://www.phosphosite.org/homeAction.action) for
  phosphoryations
- ...

Other compilations of modifications with a focus on mass spectrometry:

- [Delta Mass](https://abrf.org/resources/delta-mass/)
- [FindMod](https://web.expasy.org/findmod/)
- ...

# Different types of PTM annotations

PTMs can be annotated using different syntaxes:

- The *deltaMass* syntax that describes the mass shift of an amino
  acid such as for example `"M[+15.994915]PEPTIDE"` for an oxidation.
- The *unimodId* that uses the unimod identifier such as for example
  `"M[UNIMOD:35]PEPTIDE"` for an oxidation.
- The modification's name, such as for example `"M[Oxidation]PEPTIDE"`.


The `convertAnnotion()` function can be used to convert between these
different syntaxes:

```{r convertAnnotion}
convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "deltaMass")
convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "unimodId")
convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "name")
```
# Add modifications to a sequence

PTMs may be added (in any of the formats mentioned above) using the
`addFixedModifications()` and `addVariableModifications()` functions.

The latter will provide all combinations of possible sequences with the given
modifications. 

```{r basicAddMod}
addFixedModifications("MPEPTIDE",
    fixedModifications = c(M = "Oxidation")
)
addVariableModifications("MPEPTIDE",
    variableModifications = c(M = -15.995, T = "Phospho")
)
```

For fixed modifications, N- and C-terminal modifications are also accepted:

```{r Nterm}
addFixedModifications("MPEPT[Phospho]IDE",
    fixedModifications = c(Nterm = 304)
)
```

Notice that for both functions, the input sequence is kept as is (keeping the
already present PTM "Phospho") whilst the notation style provided is also
untouched. Combined with `convertAnnotation()`, it is possible to convert the
sequence into the preferred notation style if need be.

For variable modifications, it is possible to limit the number of possible
combinations using the `maxMods` parameter:

```{r maxMods}
addVariableModifications("MPEPTIDE",
    variableModifications = c(M = -15.995, T = "Phospho"),
    maxMods = 1
)
```

Both fixed and variable modifications can be called at once using
`addModifications()`. Use the `convertToStyle` parameter to convert the final
sequence into your preferred notation style if need be:

```{r addMod}
addModifications("MPEPTIDE",
    fixedModifications = c(M = "Oxidation"),
    variableModifications = c(T = 79.966),
    convertToStyle = "unimodId"
)
```

To remove any and all modifications on a sequence, use
`getCanonicalSequence()`:

```{r getCanonicalSeq}
getCanonicalSequence("[+304]-M[Oxidation]PEPT[UNIMOD:21]IDE")
```

# Session information

```{r si}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{elements}
\alias{elements}
\title{Elements data set.}
\format{
A `data.frame` with 4 columns (Name, FullName, AvgMass, MonoMass) for
the chemical elements.
}
\source{
Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
}
\usage{
data("elements")
}
\description{
`data.frame` of chemical elements from the unimod database.
}
\details{
It was created as
follows:

```
PTMods:::.createDataSets()
```
}
\examples{
data(elements)
head(elements)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-fixed.R
\name{addFixedModifications}
\alias{addFixedModifications}
\title{Add fixed modifications}
\usage{
addFixedModifications(sequences, fixedModifications = c(C = 57.021464))
}
\arguments{
\item{sequences}{Character vector. Peptide sequences that may have
modifications or not.}

\item{fixedModifications}{Named `numeric` or `character`. If a
`character` is given, values must be in UniMod name or UniMod ID format
(e.g. `"Phospho"`, `"UNIMOD:21"`). The annotation style of the values is
preserved in the output.
Specifies which fixed modifications are applied to which amino acids. By
default, carbamidomethylation is applied (+57.021464 on cysteines).
Supplying multiple modifications for the same amino acid is not supported;
call the function consecutively instead:
\preformatted{
  "ATK" |>
    addFixedModifications(c(T = "+57.024")) |>
    addFixedModifications(c(T = "Phospho"))
}}
}
\value{
A named list of character strings with all fixed modifications
applied to the corresponding amino acids, one element per input sequence.
}
\description{
Applies fixed modifications to peptide sequences.

The annotation style of modifications already present in `sequences` is
preserved as-is.  The annotation style of newly applied modifications
matches the style supplied in `fixedModifications`: numeric values produce
deltaMass notation (e.g. `[+79.966331]`), while character values are used
verbatim (e.g. `[Phospho]` or `[UNIMOD:21]`).
}
\examples{
addFixedModifications("ATCK")
addFixedModifications("ATCK", fixedModifications = c(Nterm = 304, C = 57.02146))
addFixedModifications("ATSK", fixedModifications = c(A = -42, S = 79.966331))
addFixedModifications("ATK", fixedModifications = c(T = "Phospho", A = 42))
addFixedModifications("A[+42]TK", fixedModifications = c(T = "Phospho"))
addFixedModifications("AT[+79.966]AK", fixedModifications = c(A = 42))
addFixedModifications("ATA[+79.966]K", fixedModifications = c(A = 42))

## Apply two modifications to the same residue consecutively
"ATK" |>
    addFixedModifications(c(T = "+57.024")) |>
    addFixedModifications(c(T = "Phospho"))
}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-variable.R
\name{addVariableModifications}
\alias{addVariableModifications}
\title{Add variable modifications}
\usage{
addVariableModifications(
  sequences,
  variableModifications = NULL,
  maxMods = Inf
)
}
\arguments{
\item{sequences}{Character vector. Peptide sequences that may have
modifications or not.}

\item{variableModifications}{Named `numeric` or `character`. If a
`character` is given, values must be in UniMod name or UniMod ID format
(e.g. `"Phospho"`, `"UNIMOD:21"`). The annotation style of the values is
preserved in the output.
Specifies which variable modifications are used on which amino acids.
Supplying multiple modifications for the same amino acid is not supported;
call the function consecutively instead:
\preformatted{
  seq |>
    addVariableModifications(c(T = "+57.024")) |>
    addVariableModifications(c(T = "Phospho"))
}}

\item{maxMods}{Numeric. Indicates how many modifications can be applied
at once.}
}
\value{
A named list of character vectors with all possible combinations
of modifications, one element per input sequence.
}
\description{
Generates all possible combinations of variable modifications
for peptide sequences.

The annotation style of modifications already present in `sequences` is
preserved as-is.  The annotation style of newly applied modifications
matches the style supplied in `variableModifications`: numeric values
produce deltaMass notation (e.g. `[+79.966331]`), while character values
are used verbatim (e.g. `[Phospho]` or `[UNIMOD:21]`).
}
\examples{
addVariableModifications(
    "APGHKA",
    variableModifications = c(A = 4, K = 5, S = 8),
    maxMods = 2
)
addVariableModifications(
    "APGHKA",
    variableModifications = c(A = 4, K = 5, S = 8),
    maxMods = 3
)
addVariableModifications(
    c("ATK", "PAK"),
    variableModifications = c(T = "Phospho", A = "UNIMOD:1")
)
addVariableModifications(
    "A[+10]KT[-5]",
    variableModifications = c(A = -20, T = "Phospho")
)

## Apply two modifications to the same residue consecutively
"ATK" |>
    addVariableModifications(c(T = "+57.024")) |>
    addVariableModifications(c(T = "Phospho"))
}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-combine.R
\name{combineModifications}
\alias{combineModifications}
\title{Combines consecutive modifications on amino acids in peptide
sequences.}
\usage{
combineModifications(sequences)
}
\arguments{
\item{sequences}{`character`. A character vector of peptide sequences with
modifications in any format accepted by `convertAnnotation`.
Multiple modifications per amino acid are allowed.}
}
\value{
A character vector of the same length as `sequences` with
consecutive modifications on each amino acid summed into a single
modification. This summed modifications is written in delta mass format.
}
\description{
Combines consecutive modifications on amino acids in peptide
sequences.
}
\examples{
combineModifications("AT[+10][-5]K")
combineModifications("[+10]-[-5]-ATK")
combineModifications("AT[Phospho][UNIMOD:1]K")
combineModifications("EM[+15.9949][-10]EK[+42][-8]")
combineModifications(c("AT[+10][-5]K", "EM[+15.9949][-10]EK"))

}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-utils.R
\name{getModificationsCounts}
\alias{getModificationsCounts}
\title{Extract modification counts from one or more peptide sequences}
\usage{
getModificationsCounts(sequences)
}
\arguments{
\item{sequences}{A `character()` vector of peptide sequences in ProForma
format.}
}
\value{
A named integer vector where names are UniMod modification names
and values are their occurrence counts across all input sequences.
Returns an empty integer vector if no modifications are found.
}
\description{
Given a character vector of peptide sequences containing modifications in
any annotation style (deltaMass, UniMod ID, or name), returns a named
integer vector with the count of each unique modification. Regardless of
the input annotation style, all modifications are converted to their
UniMod name in the output (e.g. `[+79.966]` and `[UNIMOD:21]` both
appear as `Phospho`).
}
\examples{
seqs <- c(
    "[+304]-AT[Phospho]K",
    "AC[Carbamidomethyl]T[+79.966]S[Phospho]K",
    "PEPTIDE"
)
getModificationsCounts(seqs)

}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{modifications}
\alias{modifications}
\title{Modifications data set.}
\format{
A `data.frame` with 15 columns (Id (created by
Name:(Position-)Site(:NeutralLoss) because unimod id is not unique),
UnimodId, Name, Description, AvgMass, MonoMass, Site, Position,
Classification, SpecGroup, NeutralLoss, LastModification, Approved,
Hidden) for the modifications.
}
\source{
Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
}
\usage{
data("modifications")
}
\description{
`data.frame` of modifications from the unimod database.
}
\details{
It was created as
follows:

```
PTMods:::.createDataSets()
```
}
\examples{
data(modifications)
head(modifications)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-add.R
\name{addModifications}
\alias{addModifications}
\title{Add fixed and variable modifications to peptide sequences}
\usage{
addModifications(
  sequences,
  fixedModifications = NULL,
  variableModifications = NULL,
  convertToStyle = NULL,
  ...
)
}
\arguments{
\item{sequences}{`character`. Peptide sequences with or without existing
modification annotations. Modifications must be enclosed in square
brackets (e.g. `"AT[+79.966]K"`, `"[+304]-ATK"`).}

\item{fixedModifications}{Named `numeric` or `character`, or `NULL`.
Fixed modifications to be applied unconditionally to the specified amino
acids or termini. Names correspond to single-letter amino acid codes,
`"Nterm"`, or `"Cterm"`. If `character`, values must be in UniMod name
or UniMod ID format (e.g. `"Phospho"`, `"UNIMOD:21"`). See
[addFixedModifications()] for full details. Set to `NULL` to skip.}

\item{variableModifications}{Named `numeric` or `character`, or `NULL`.
Variable modifications to enumerate over the specified amino acids.
Names correspond to single-letter amino acid codes. If `character`,
values must be in UniMod name or UniMod ID format. See
[addVariableModifications()] for full details. Set to `NULL` to skip.}

\item{convertToStyle}{`character(1)`. Controls the annotation format of
modifications in the returned sequences. One of:
\describe{
  \item{`NULL`}{(default) Same as input modification style.}
  \item{`"deltaMass"`}{Delta mass notation, e.g. `[+79.966]`.}
  \item{`"unimodId"`}{UniMod ID notation, e.g. `[UNIMOD:21]`.}
  \item{`"name"`}{UniMod name notation, e.g. `[Phospho]`.}
}
Conversion is performed by [convertAnnotation()].}

\item{...}{Additional arguments passed to the underlying modification
functions. Currently supported:
\describe{
  \item{`maxMods`}{Forwarded to [addVariableModifications()]. A
    `numeric(1)` specifying the maximum number of variable modifications
    that may be applied simultaneously. Defaults to `Inf`.}
}}
}
\value{
A `character` vector of peptidoforms. The length is greater than or
  equal to `length(sequences)`, as each input sequence may expand into
  multiple peptidoforms corresponding to all combinations of variable
  modifications.
}
\description{
A convenience wrapper around [addFixedModifications()] and
[addVariableModifications()] that applies both fixed and variable
modifications in a single call. Fixed modifications are applied first,
followed by the enumeration of all possible combinations of variable
modifications. Optionally, the annotation style of
the output can be converted via [convertAnnotation()].
}
\examples{
## Fixed modifications only
addModifications(
    c("ATCK", "PEPCK"),
    fixedModifications = c(C = 57.02146)
)

## Variable modifications only
addModifications(
    c("ATSK", "PEPTSK"),
    variableModifications = c(S = 79.966331, T = 79.966331),
    maxMods = 2
)

## Both fixed and variable modifications
addModifications(
    "ATCSK",
    fixedModifications = c(C = 57.02146),
    variableModifications = c(S = 79.966331, T = 79.966331),
    maxMods = 1
)

## Return annotations in UniMod name style
addModifications(
    "ATSK",
    variableModifications = c(S = 79.966331, T = 79.966331),
    convertToStyle = "name"
)

## N-terminal fixed modification with variable modifications
addModifications(
    "ATSK",
    fixedModifications = c(Nterm = 304),
    variableModifications = c(S = 79.966331)
)
}
\seealso{
* \code{\link{addFixedModifications}} for applying fixed modifications.

* \code{\link{addVariableModifications}} for enumerating variable modification
  combinations.

* \code{\link{convertAnnotation}} for converting between modification annotation
  styles.
}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PTM-utils.R
\name{getCanonicalSequence}
\alias{getCanonicalSequence}
\title{Get canonical sequence from any peptidoform}
\usage{
getCanonicalSequence(x)
}
\arguments{
\item{x}{`character`, ProForma sequence.}
}
\value{
`character`, a `character` vector with sequences cleaned of all
modifications.
}
\description{
Remove any modifications annotation on a sequence. Annotation
from any type (so long as they are specified within brackets "[..]") are
removed.
}
\examples{
getCanonicalSequence(
    c(
        "EM[-15.9949]EVEES[+79.9663]PEK",
        "EK[Acetyl]EVEES[UNIMOD:21]PEK"
    )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{PTMods-package}
\alias{PTMods}
\alias{PTMods-package}
\title{Managing amino acid modifications for mass spectrometry in R.}
\description{
The `PTMods` package focuses on handling post-translational modifications
(PTMs). It distributes PTMs from the Unimod database and allows to convert
PTM annotations between multiple formats.
}
\references{
\url{https://github.com/RforMassSpectrometry/PTMods/}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/RforMassSpectrometry/PTMods}
  \item Report bugs at \url{https://github.com/RforMassSpectrometry/PTMods/issues}
}

}
\author{
\strong{Maintainer}: Guillaume Deflandre \email{guillaume.deflandre@uclouvain.be} (\href{https://orcid.org/0009-0008-1257-2416}{ORCID})

Authors:
\itemize{
  \item Laurent Gatto \email{laurent.gatto@uclouvain.be} (\href{https://orcid.org/0000-0002-1520-2268}{ORCID})
  \item Sebastian Gibb \email{mail@sebastiangibb.de} (\href{https://orcid.org/0000-0001-7406-4443}{ORCID})
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertAnnotation.R
\name{convertAnnotation}
\alias{convertAnnotation}
\alias{.convertAnnotation}
\title{Convert sequences from one annotation style to another}
\usage{
convertAnnotation(
  x,
  convertToStyle = c("deltaMass", "unimodId", "name"),
  massTolerance = 0.01,
  verbose = TRUE
)
}
\arguments{
\item{x}{Character vector with peptide sequences in ProForma format}

\item{convertToStyle}{Character string specifying target format. Options:
"deltaMass", "unimodId", "name"}

\item{massTolerance}{Numeric mass tolerance in Daltons for matching
modifications (default: 0.01). Used when converting from deltaMass.}

\item{verbose}{`logical(1)`. If `FALSE`, warnings about unrecognised
modifications are silenced (default: `TRUE`).}
}
\value{
Character vector with the sequences in the target annotation format
}
\description{
Converts modifications between different annotation formats for multiple
sequences at once. See the details and examples sections for more
information.  The annotation styles are inferred from the `modifications`
dataframe (see `?modifications`).
}
\details{
The function handles three main conversion scenarios:

\itemize{
  \item Name to deltaMass: \code{"M[Oxidation]PEPTIDE"} -> \code{"M[+15.994915]PEPTIDE"}
  \item Name to unimodId: \code{"M[Oxidation]PEPTIDE"} -> \code{"M[UNIMOD:35]PEPTIDE"}
  \item deltaMass to name: \code{"M[+15.995]PEPTIDE"} -> \code{"M[Oxidation]PEPTIDE"}
  \item deltaMass to unimodId: \code{"M[+15.995]PEPTIDE"} -> \code{"M[UNIMOD:35]PEPTIDE"}
  \item unimodId to name: \code{"M[UNIMOD:35]PEPTIDE"} -> \code{"M[Oxidation]PEPTIDE"}
  \item unimodId to deltaMass: \code{"M[UNIMOD:35]PEPTIDE"} -> \code{"M[+15.994915]PEPTIDE"}
}
}
\examples{
# Convert sequence from name to delta mass
convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "deltaMass")
# Result: "M[+15.994915]PEPTIDE"

# Name to Unimod ID
convertAnnotation("M[Oxidation]PEPTIDE", convertToStyle = "unimodId")
# Result: "M[UNIMOD:35]PEPTIDE"

# Delta mass to name
convertAnnotation("M[+15.995]PEPTIDE", convertToStyle = "name")
# Result: "M[Oxidation]PEPTIDE"

# Multiple modifications
convertAnnotation("M[Oxidation]EVNES[Phospho]PEK", convertToStyle = "deltaMass")
# Result: "M[+15.994915]EVNES[+79.966331]PEK"

# Convert multiple sequences from name to delta mass
sequences <- c("M[Oxidation]PEPTIDE", "EVNES[Phospho]PEK", "PEPTIDE")
convertAnnotation(sequences, convertToStyle = "deltaMass")
# Result: c("M[+15.994915]PEPTIDE", "EVNES[+79.966331]PEK", "PEPTIDE")

# Convert from delta mass to name
sequences <- c("M[+15.995]PEPTIDE", "S[+79.966]EQUENCE")
convertAnnotation(sequences, convertToStyle = "name")
# Result: c("M[Oxidation]PEPTIDE", "S[Phospho]EQUENCE")

# Convert to Unimod IDs
sequences <- c("M[Oxidation]PEPTIDE", "C[Carbamidomethyl]PEPTIDE")
convertAnnotation(sequences, convertToStyle = "unimodId")
# Result: c("M[UNIMOD:35]PEPTIDE", "C[UNIMOD:4]PEPTIDE")

}
\author{
Guillaume Deflandre <guillaume.deflandre@uclouvain.be>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{aminoacids}
\alias{aminoacids}
\title{Amminoacids data set.}
\format{
A `data.frame` with 11 columns (OneLetter, ThreeLetter, FullName,
AvgMass, MonoMass, H, C, N, O S, Se) for the aminoacids. The H/C/N/O/S/Se
columns contain the number of elements that build the aminoacid.
}
\source{
Taken from the unimod database: http://www.unimod.org/xml/unimod.xml.
}
\usage{
data("aminoacids")
}
\description{
`data.frame` of aminoacids from the unimod database.
}
\details{
It was created as
follows:

```
PTMods:::.createDataSets()
```
}
\examples{
data(aminoacids)
head(aminoacids)
}
\keyword{datasets}
